#!/usr/bin/env python
# vim: set expandtab ts=4 sw=4:

# Copyright (C) 2009 University of Victoria
# You may distribute under the terms of either the GNU General Public
# License or the Apache v2 License, as specified in the README file.

## Auth: Patrick Armstrong. 8/28/2009.
##
## Cloud Scheduler Information Server
## This class implements an XMLRPC Server that serves information about the state
## of the cloud sceduler to information utilities (web interface, command line, whatever)
##
## Based on http://docs.python.org/library/simplexmlrpcserver.html

##
## IMPORTS
##
import logging
import threading
import time
import socket
import sys
import platform
import re
from SimpleXMLRPCServer import SimpleXMLRPCServer
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler

import cloudscheduler.config as config
from cluster_tools import ICluster
from cluster_tools import VM
from cloud_management import ResourcePool
from job_management import Job
from job_management import JobPool
# JSON lib included in 2.6+
if sys.version_info < (2, 6):
    try:
        import simplejson as json
    except:
        raise "Please install the simplejson lib for python 2.4 or 2.5"
else:
    import json

log = None

class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ('/RPC2',)

class InfoServer(threading.Thread,):

    cloud_resources = None
    job_pool = None
    job_poller = None
    machine_poller = None
    vm_poller = None
    scheduler = None
    cleaner = None
    def __init__(self, c_resources, c_job_pool, c_job_poller, c_machine_poller, c_vm_poller, c_scheduler, c_cleaner):

        global log
        log = logging.getLogger("cloudscheduler")

        #set up class
        threading.Thread.__init__(self, name=self.__class__.__name__)
        self.done = False
        cloud_resources = c_resources
        job_pool = c_job_pool
        job_poller = c_job_poller
        machine_poller = c_machine_poller
        vm_poller = c_vm_poller
        scheduler = c_scheduler
        cleaner = c_cleaner
        host_name = "0.0.0.0"
        #set up server
        try:
            self.server = SimpleXMLRPCServer((host_name,
                                              config.info_server_port),
                                              requestHandler=RequestHandler,
                                              logRequests=False)
            self.server.socket.settimeout(1)
            self.server.register_introspection_functions()
        except:
            log.error("Couldn't start info server: %s" % sys.exc_info()[0])
            sys.exit(1)

        # Register an instance; all the methods of the instance are
        # published as XML-RPC methods
        class externalFunctions:
            def get_cloud_resources(self):
                return cloud_resources.get_pool_info()
            def get_cluster_resources(self):
                output = []
                output.append("Clusters in resource pool:\n")
                for cluster in cloud_resources.resources:
                    output.append(cluster.get_cluster_info_short())
                    output.append("\n")
                return ''.join(output)
            def get_cluster_vm_resources(self):
                output = []
                output.append(VM.get_vm_info_header())
                clusters = 0
                vm_count = 0
                for cluster in cloud_resources.resources:
                    clusters += 1
                    vm_count += len(cluster.vms)
                    output.append(cluster.get_cluster_vms_info())
                output.append('\nTotal VMs: %i. Total Clouds: %i' % (vm_count, clusters))
                extra_output = []
                for cluster in cloud_resources.retired_resources:
                    extra_output.append(cluster.get_cluster_vms_info())
                if len(extra_output) > 0:
                    output.append('\n\nRetiring VMs from removing resources:\n')
                    output.extend(extra_output)
                return ''.join(output)
            def get_cluster_info(self, cluster_name):
                output = []
                output.append("Cluster Info: %s\n" % cluster_name)
                cluster = cloud_resources.get_cluster(cluster_name)
                if cluster:
                    output.append(cluster.get_cluster_info_short())
                else:
                    output.append("Cluster named %s not found." % cluster_name)
                return ''.join(output)
            def get_vm_info(self, cluster_name, vm_id):
                output = []
                output.append("VM Info for VM id: %s\n" % vm_id)
                cluster = cloud_resources.get_cluster(cluster_name)
                vm = None
                if cluster:
                    vm = cluster.get_vm(vm_id)
                else:
                    output.append("Cluster %s not found.\n" % cluster_name)
                if vm:
                    output.append(vm.get_vm_info())
                else:
                    output.append("VM with id: %s not found.\n" % vm_id)
                return ''.join(output)
            def get_json_vm(self, cluster_name, vm_id):
                output = "{}"
                cluster = cloud_resources.get_cluster(cluster_name)
                vm = None
                if cluster:
                    vm = cluster.get_vm(vm_id)
                    if vm:
                        output = VMJSONEncoder().encode(vm)
                return output
            def get_json_cluster(self, cluster_name):
                output = "{}"
                cluster = cloud_resources.get_cluster(cluster_name)
                if cluster:
                    output = ClusterJSONEncoder().encode(cluster)
                return output
            def get_json_resource(self):
                return ResourcePoolJSONEncoder().encode(cloud_resources)
            def get_developer_information(self):
                try:
                    from guppy import hpy
                    h = hpy()
                    heap = h.heap()
                    return str(heap)
                except:
                    return "You need to have Guppy installed to get developer " \
                           "information" 
            def get_newjobs(self):
                output = []
                jobs = job_pool.job_container.get_unscheduled_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_schedjobs(self):
                output = []
                jobs = job_pool.job_container.get_scheduled_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_highjobs(self):
                output = []
                jobs = job_pool.job_container.get_high_priority_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_idlejobs(self):
                output = []
                jobs = job_pool.job_container.get_idle_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_runningjobs(self):
                output = []
                jobs = job_pool.job_container.get_running_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_completejobs(self):
                output = []
                jobs = job_pool.job_container.get_complete_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_heldjobs(self):
                output = []
                jobs = job_pool.job_container.get_held_jobs()
                output.append(Job.get_job_info_header())
                for job in jobs:
                    output.append(job.get_job_info())
                return ''.join(output)
            def get_job(self, jobid):
                output = "Job not found."
                job = job_pool.job_container.get_job_by_id(jobid)
                if job != None:
                    output = job.get_job_info_pretty()
                return output
            def get_json_job(self, jobid):
                job_match = job_pool.job_container.get_job_by_id(jobid)
                return JobJSONEncoder().encode(job_match)
            def get_json_jobpool(self):
                return JobPoolJSONEncoder().encode(job_pool)
            def get_ips_munin(self):
                output = []
                for cluster in cloud_resources.resources:
                    for vm in cluster.vms:
                        if re.search("(10|192\.168|172\.(1[6-9]|2[0-9]|3[01]))\.", vm.ipaddress):
                            continue
                        else:
                            output.append("[%s]\n\taddress %s\n" % (vm.hostname, vm.ipaddress))
                return ''.join(output)
            def get_vm_startup_time(self):
                output = []
                for cluster in cloud_resources.resources:
                    output.append("Cluster: %s " % cluster.name)
                    total_time = 0
                    for vm in cluster.vms:
                        pass
                        output.append("%d, " % (vm.startup_time if vm.startup_time != None else 0))
                        total_time += (vm.startup_time if vm.startup_time != None else 0)
                    if len(cluster.vms) > 0:
                        output.append(" Avg: %d " % (int(total_time) / len(cluster.vms)))
                return ''.join(output)
            def get_diff_types(self):
                output = []
                current_types = cloud_resources.vmtype_distribution()
                desired_types = job_pool.job_type_distribution()
                # Negative difference means will need to create that type
                diff_types = {}
                for vmtype in current_types.keys():
                    if vmtype in desired_types.keys():
                        diff_types[vmtype] = current_types[vmtype] - desired_types[vmtype]
                    else:
                        diff_types[vmtype] = 1 # changed from 0 to handle users with multiple job types, back to 0 from 1.
                for vmtype in desired_types.keys():
                    if vmtype not in current_types.keys():
                        diff_types[vmtype] = -desired_types[vmtype]
        
                # With user limiting will need to reset any users that are at their limits
                # so they will not interfere with scheduling
                # will need to redistribute negatives to the non-limited users
                limited_users = []
                userjoblimits = job_pool.get_usertype_limits()
                for vmusertype in diff_types.keys():
                    user = vmusertype.split(':')[0]
                    if cloud_resources.user_at_limit(user):
                        if vmusertype not in limited_users:
                            limited_users.append(vmusertype)
                    if vmusertype in userjoblimits.keys():
                        if cloud_resources.uservmtype_at_limit(vmusertype, userjoblimits[vmusertype]):
                            if vmusertype not in limited_users:
                                limited_users.append(vmusertype)
                neg_total = 0
                for usertype in limited_users:
                    if diff_types[usertype] < 0:
                        neg_total += diff_types[usertype]
                splitby = len(diff_types) - len(limited_users)
                adjustby = 0
                if splitby > 0:
                    adjustby = neg_total / splitby
                elif splitby == 0:
                    log.verbose("All users are limited.")
                else:
                    log.error("More user vmtypes limited than what's in diff types, something weird here.")
        
                for usertype in diff_types.keys():
                    if usertype not in limited_users:
                        diff_types[usertype] += adjustby # the 'extra' will be negative so add it
                        
                #for type in current_types.keys():
                    #if type in desired_types.keys():
                        #diff_types[type] = current_types[type] - desired_types[type]
                    #else:
                        #diff_types[type] = 1 # changed from 0 to handle users with multiple job types
                #for type in desired_types.keys():
                    #if type not in current_types.keys():
                        #diff_types[type] = -desired_types[type]
                output.append("Diff Types dictionary\n")
                for key, value in diff_types.iteritems():
                    output.append("type: %s, dist: %f\n" % (key, value))
                output.append("Current Types (vms)\n")
                for key, value in current_types.iteritems():
                    output.append("type: %s, dist: %f\n" % (key, value))
                output.append("Desired Types (jobs)\n")
                for key, value in desired_types.iteritems():
                    output.append("type: %s, dist: %f\n" % (key, value))
                return ''.join(output)

            def get_vm_job_run_times(self):
                output = []
                output.append("Run Times of Jobs on VMs\n")
                for cluster in cloud_resources.resources:
                    for vm in cluster.vms:
                        output.append("%s : avg %f\n" % (vm.hostname, vm.job_run_times.average()))
                return ''.join(output)
            def get_cloud_config_values(self):
                return cloud_resources.get_cloud_config_output()
            def get_total_vms(self):
                return str(cloud_resources.vm_count())
            def get_total_vms_cloud(self, cluster_name):
                output = []
                cluster = cloud_resources.get_cluster(cluster_name)
                if cluster:
                    output.append(str(cluster.num_vms()))
                else:
                    output.append("Cluster named %s not found." % cluster_name)
                return ''.join(output)
            def check_shared_objs(self):
                output = []
                output.append("Scheduler Thread:\n" + scheduler.check_shared_objs())
                output.append("\n")
                output.append("Cleanup Thread:\n" + cleaner.check_shared_objs())
                output.append("\n")
                output.append("VMPoller Thread:\n" + vm_poller.check_shared_objs())
                output.append("\n")
                output.append("JobPoller Thread:\n" + job_poller.check_shared_objs())
                output.append("\n")
                output.append("MachinePoller Thread:\n" + machine_poller.check_shared_objs())
                output.append("\n")
                return ''.join(output)
            def get_vm_stats(self, cluster_name=""):
                vms = []
                if cluster_name:
                    cluster = cloud_resources.get_cluster(cluster_name)
                    if not cluster:
                        return "Could not find cloud: %s - check cloud_status for list of available clouds" % cluster_name
                    vms.extend(cluster.vms)
                else:
                    for cluster in cloud_resources.resources:
                        vms.extend(cluster.vms)
                state_count = {'Running':0, 'Starting':0, 'Error':0, 'Retiring':0, 'ExpiredProxy':0, 'NoProxy':0, 'ConnectionRefused':0}
                for vm in vms:
                    if vm.override_status:
                        state_count[vm.override_status] += 1
                    else:
                        state_count[vm.status] += 1
                output = []
                for state, count in state_count.iteritems():
                    if count > 0:
                        output.extend([str(count), ' ', state, ','])
                return ''.join(output)
            def get_missing_vm_list(self):
                return cloud_resources.fetch_missing_vm_list()
            def get_thread_heart_beats(self):
                now = time.time()
                output = []
                output.append("Thread Heart beat times:\n")
                output.append("   Scheduler Thread(%s): %s\n" % (scheduler.scheduling_interval, str(int(now - scheduler.heart_beat))))
                output.append("   Cleanup Thread(%s): %s\n" % (cleaner.polling_interval, str(int(now - cleaner.heart_beat))))
                output.append("   VMPoller Thread(%s): %s\n" % (vm_poller.run_interval, str(int(now - vm_poller.heart_beat))))
                output.append("   JobPoller Thread(%s): %s\n" % (job_poller.polling_interval, str(int(now - job_poller.heart_beat))))
                output.append("   MachinePoller Thread(%s): %s\n" % (machine_poller.polling_interval, str(int(now - machine_poller.heart_beat))))
                return ''.join(output)
            def get_job_failure_reasons(self):
                output = []
                output.append("Job Failure Reasons:\n")
                reasons = job_pool.fetch_job_failure_reasons()
                for job in reasons:
                    output.append("   Job ID: %s\n" % job.id)
                    for reason in job.failed_boot_reason:
                        output.append("      %s\n" % reason)
                return ''.join(output)
            def get_image_failures(self):
                output = []
                output.append("Image Failure List\n")
                for cloud in cloud_resources.resources:
                    if len(cloud.failed_image_set) > 0:
                        output.append("   Cloud: %s\n" % cloud.name)
                        for image in cloud.failed_image_set:
                            output.append("      Image: %s\n" % image)
                return ''.join(output)

        self.server.register_instance(externalFunctions())

    def run(self):

        # Run the server's main loop
        log.info("Started info server on port %s" % config.info_server_port)
        while self.server:
            try:
                self.server.handle_request()
                if self.done:
                    log.debug("Killing info server...")
                    self.server.socket.close()
                    break
            except socket.timeout:
                log.warning("info server's socket timed out. Don't panic!")

    def stop(self):
        self.done = True

class VMJSONEncoder(json.JSONEncoder):
    def default(self, vm):
        if not isinstance (vm, VM):
            log.error("Cannot use VMJSONEncoder on non VM object of type %s, %s" % (type(vm), vm))
            return
        return {'name': vm.name, 'id': vm.id, 'vmtype': vm.vmtype,
                'hostname': vm.hostname, 'clusteraddr': vm.clusteraddr,
                'ipaddress': vm.ipaddress,
                'cloudtype': vm.cloudtype, 'network': vm.network, 
                'image': vm.image, 'alt_hostname': vm.alt_hostname,
                'memory': vm.memory, 'mementry': vm.mementry, 'flavor': vm.flavor,
                'cpucores': vm.cpucores, 'storage': vm.storage, 
                'status': vm.status, 'condoraddr': vm.condoraddr,
                'condorname': vm.condorname, 'condormasteraddr': vm.condormasteraddr,
                'keep_alive': vm.keep_alive, 'user': vm.user, 'uservmtype': vm.uservmtype,
                'clusteraddr': vm.clusteraddr, 'clusterport': vm.clusterport,
                'errorcount': vm.errorcount, 'errorconnect': vm.errorconnect,
                'lastpoll': vm.lastpoll, 'last_state_change': vm.last_state_change,
                'initialize_time': vm.initialize_time, 'startup_time': vm.startup_time,
                'idle_start': vm.idle_start, 'spot_id': vm.spot_id,
                'proxy_file': vm.proxy_file, 'myproxy_creds_name': vm.myproxy_creds_name,
                'myproxy_server': vm.myproxy_server, 'myproxy_server_port': vm.myproxy_server_port,
                'myproxy_renew_time': vm.myproxy_renew_time, 'override_status': vm.override_status,
                'job_per_core': vm.job_per_core, 'force_retire': vm.force_retire,
                'failed_retire': vm.failed_retire, 'x509userproxy_expiry_time': str(vm.x509userproxy_expiry_time),
                'job_run_times': vm.job_run_times.data}

class ClusterJSONEncoder(json.JSONEncoder):
    def default(self, cluster):
        if not isinstance (cluster, ICluster):
            log.error("Cannot use ClusterJSONEncoder on non Cluster object")
            return
        vmEncodes = []
        for vm in cluster.vms:
            vmEncodes.append(VMJSONEncoder().encode(vm))
        vmDecodes = []
        for vm in vmEncodes:
            vmDecodes.append(json.loads(vm))
        return {'name': cluster.name, 'network_address': cluster.network_address,
                'cloud_type': cluster.cloud_type, 'memory': cluster.memory, 
                'network_pools': cluster.network_pools, 
                'vm_slots': cluster.vm_slots, 'cpu_cores': cluster.cpu_cores, 
                'storageGB': cluster.storageGB, 'vms': vmDecodes, 'enabled':cluster.enabled,
                'hypervisor': cluster.hypervisor, 'max_mem': cluster.max_mem,
                'max_vm_mem': cluster.max_vm_mem, 'max_slots': cluster.max_slots,
                'max_storageGB': cluster.max_storageGB, 'boot_timeout': cluster.boot_timeout,
                'connection_fail_disable_time': cluster.connection_fail_disable_time,
                'connection_problem': cluster.connection_problem,
                'errorconnect': cluster.errorconnect}

class ResourcePoolJSONEncoder(json.JSONEncoder):
    def default(self, res_pool):
        if not isinstance (res_pool, ResourcePool):
            log.error("Cannot use ResourcePoolJSONEncoder on non ResourcePool Object")
            return
        pool = []
        for cluster in res_pool.resources:
            pool.append(ClusterJSONEncoder().encode(cluster))
        poolDecodes = []
        for cluster in pool:
            poolDecodes.append(json.loads(cluster))
        return {'resources': poolDecodes}

class JobJSONEncoder(json.JSONEncoder):
    def default(self, job):
        if not isinstance(job, Job):
            log.error("Cannot use JobJSONEncoder on non Job Object")
            return
        return {'id': job.id, 'user': job.user, 'priority': job.priority,
                'job_status': job.job_status, 'cluster_id': job.cluster_id,
                'proc_id': job.proc_id, 'req_vmtype': job.req_vmtype,
                'req_network': job.req_network,
                'req_image': job.req_image, 'req_imageloc': job.req_imageloc,
                'req_ami': job.req_ami, 'req_memory': job.req_memory,
                'req_cpucores': job.req_cpucores, 'req_storage': job.req_storage,
                'keep_alive': job.keep_alive, 'status': job.status,
                'remote_host': job.remote_host, 'running_cloud': job.running_cloud,
                'banned': job.banned, 'ban_time': job.ban_time, 'target_clouds':job.target_clouds,
                'blocked_clouds':job.blocked_clouds, 'uservmtype': job.uservmtype,
                'high_priority': job.high_priority, 'instance_type': job.instance_type,
                'maximum_price': job.maximum_price, 'spool_dir': job.spool_dir,
                'myproxy_server': job.myproxy_server, 'myproxy_server_port': job.myproxy_server_port,
                'myproxy_creds_name': job.myproxy_creds_name, 'running_vm': job.running_vm,
                'x509userproxysubject': job.x509userproxysubject, 'x509userproxy': job.x509userproxy,
                'original_x509userproxy': job.original_x509userproxy,
                'x509userproxy_expiry_time': job.x509userproxy_expiry_time,
                'proxy_renew_time': job.proxy_renew_time, 'job_per_core': job.job_per_core,
                'servertime': job.servertime, 'jobstarttime': job.jobstarttime,
                'machine_reserved': job.machine_reserved, 'req_hypervisor': job.req_hypervisor,
                'proxy_non_boot': job.proxy_non_boot, 'vmimage_proxy_file': job.vmimage_proxy_file,
                'usertype_limit': job.usertype_limit, 'req_image_id': job.req_image_id,
                'req_instance_type_ibm': job.req_instance_type_ibm, 'location': job.location,
                'key_name': job.key_name, 'req_security_group': job.req_security_group,
                'override_status': job.override_status, 'block_time': job.block_time
                }

class JobPoolJSONEncoder(json.JSONEncoder):
    def default(self, job_pool):
        if not isinstance(job_pool, JobPool):
            log.error("Cannot use JobPoolJSONEncoder on non JobPool Object")
            return
        new_queue = []
        for job in job_pool.job_container.get_unscheduled_jobs():
            new_queue.append(JobJSONEncoder().encode(job))
        sched_queue = []
        for job in job_pool.job_container.get_scheduled_jobs():
            sched_queue.append(JobJSONEncoder().encode(job))
        new_decodes = []
        for job in new_queue:
            print job
            new_decodes.append(job)
        sched_decodes = []
        for job in sched_queue:
            sched_decodes.append(job)
        return {'new_jobs': new_decodes, 'sched_jobs': sched_decodes}



import cluster_tools
import logging
import threading

class IBMCluster(cluster_tools.ICluster):

    VM_STATES = ['Running', 'Rebooting', 'Terminated', 'Pending', 'Unknown']
    VM_STATES_DICT = {'Running': 0, 'Rebooting': 1, 'Terminated': 2, 
                      'Pending': 3, 'Unknown': 4}
    VM_COMPUTE_SIZE_MAP = { 'brz32': 0, 'bronze32': 0,
                            'brz64': 1, 'bronze64': 1,
                            'cop32': 2, 'copper32': 2,
                            'cop64': 3, 'copper64': 3, 
                            'slv32': 4, 'silver32': 4,
                            'slv64': 5, 'silver64': 5,
                            'gld32': 6, 'gold32': 6,
                            'gld64': 7, 'gold64': 7,
                        'plt64': 8, 'platinum64': 8 }

    CLOUD_LOCATION_ID_MAP = {'raleigh': '41', 'ehningen': '61', 'boulder2': '81',
                             'boulder1': '82', 'markham': '101', 'makuhari': '121',
                             'singapore': '141',}

    def __init__(self, name="Dummy Cluster", host="localhost", cloud_type="Dummy",
                 memory=[], max_vm_mem= -1, networks=[], vm_slots=0,
                 cpu_cores=0, storage=0, hypervisor='xen', username="", password="",enabled=True, priority = 0):

        # Call super class's init
        cluster_tools.ICluster.__init__(self,name=name, host=host, cloud_type=cloud_type,
                         memory=memory, max_vm_mem=max_vm_mem, networks=networks,
                         vm_slots=vm_slots, cpu_cores=cpu_cores,
                         storage=storage, hypervisor=hypervisor, enabled=enabled, priority=priority)
        from libcloud.compute.types import Provider
        from libcloud.compute.providers import get_driver
        global log
        log = logging.getLogger("cloudscheduler")
        self.username = username
        self.password = password
        self.driver = get_driver(Provider.IBM)
        self.total_cpu_cores = -1
        #self.connection = self.driver(self.username, self.password)
        #self.locations = self.connection.list_locations()
        #self.locations_dict = {loc.id: loc for loc in self.locations}
        #self.compute_sizes = self.connection.list_sizes()
        #self.images = self.connection.list_images()

    def __getstate__(self):
        """Override to work with pickle module."""
        state = self.__dict__.copy()
        del state['vms_lock']
        del state['res_lock']
        del state['driver']
        return state

    def __setstate__(self, state):
        """Override to work with pickle module."""
        self.__dict__ = state
        self.vms_lock = threading.RLock()
        self.res_lock = threading.RLock()
        self.driver = get_driver(Provider.IBM)

    def _get_connection(self, username, password):
        self.connection = self.driver(username, password)
        try:
            self.locations = self.connection.list_locations()
            self.locations_dict = {}
            for loc in self.locations:
                self.locations_dict[loc.id] = loc
            #self.locations_dict = {loc.id: loc for loc in self.locations}
            self.compute_sizes = self.connection.list_sizes()
            self.images = self.connection.list_images()
        except:
            return None
        return self.connection

    def vm_create(self, vm_name, vm_type, vm_user, vm_networkassoc,
                  vm_image, vm_mem, vm_cores, vm_storage, customization, 
                  vm_keepalive, instance_type, location, job_per_core,
                  vm_keyname, username="", password=""):
        # will need to use self.driver.deploy_node(...) as this seems to allow for contextualization whereas create_node() does not
        from libcloud.compute.base import NodeImage, NodeSize, NodeLocation, NodeAuthSSHKey
        if not username:
            username = self.username
        if not password:
            password = self.password
        conn = self._get_connection(username, password)
        if instance_type and instance_type.lower() in self.VM_COMPUTE_SIZE_MAP.keys():
            vm_size = self.compute_sizes[self.VM_COMPUTE_SIZE_MAP[instance_type.lower()]]
        else:
            log.debug("%s not a valid instance type." % instance_type)
            return
        if location.lower() in self.CLOUD_LOCATION_ID_MAP.keys():
            try:
                vm_location = self.locations_dict[self.CLOUD_LOCATION_ID_MAP[location.lower()]]
            except KeyError:
                log.debug("Bad Dict Key mapping")
                return
        else:
            log.debug("%s is not a valid location" % location)
            print 'location invalid'
            return

        image = NodeImage(vm_image, '','')
        # 20035253 image id is a RHL 5.7 that should boot on bronze 32bit and is on markham location
        instance = None
        vm_key = NodeAuthSSHKey(vm_keyname)

        instance = conn.create_node(name=vm_name, image=image, size=vm_size, location=vm_location, auth=vm_key)
        if instance:
            new_vm = VM(name = vm_name, id = instance.uuid, vmtype = vm_type, user = vm_user,
                        clusteraddr = self.network_address,
                        cloudtype = self.cloud_type, network = vm_networkassoc,
                        image= vm_image,
                        memory = vm_mem,
                        cpucores = vm_cores, storage = vm_storage, 
                        keep_alive = vm_keepalive, job_per_core = job_per_core)
            try:
                self.resource_checkout(new_vm)
            except:
                log.exception("Unexpected Error checking out resources when creating a VM. Programming error?")
                self.vm_destroy(new_vm, reason="Failed Resource checkout")
                return self.ERROR
    
            self.vms.append(new_vm)
        else:
            log.debug("No return from create_node()")

        return 0

    def vm_destroy(self, vm, return_resources=True, reason=""):
        nodes = self.connection.list_nodes()
        for node in nodes:
            if node.uuid == vm.id:
                self.connection.destroy_node(node)
                log.info("VM %s Destroyed: Reason: %s" % (vm.id, reason))
                # return resources
                if return_resources:
                    self.resource_return(vm)
                with self.vms_lock:
                    self.vms.remove(vm)
                return 0
        

    def vm_poll(self, vm):
        # libcloud does not seem to support polling individual VMs you simply list off what you have
        # ineffecient this way but is in line with other clouds
        nodes = self.connection.list_nodes()
        for node in nodes:
            if node.uuid == vm.id:
                if not vm.ipaddress:
                    if node.public_ip:
                        vm.ipaddress = node.public_ip[0]
                if self.VM_STATES[node.state] == vm.status:
                    continue
                # Both startup and shutdown enter the Pending state
                elif vm.status == 'Running' and node.state == 3:
                    vm.status = self.VM_STATES[node.state]
                    vm.override_status = 'Stopping'
                    vm.last_state_change = int(time.time())
                    log.debug("VM: %s on %s. Changed from %s to Stopping." % (vm.id, self.name, vm.status))
                elif vm.status == 'Starting' and node.state == 3:
                    vm.status = self.VM_STATES[node.state]
                    vm.last_state_change = int(time.time())
                    log.debug("VM: %s on %s. Changed from %s to %s." % (vm.id, self.name, vm.status, self.VM_STATES[node.state]))
                else:
                    vm.status = self.VM_STATES[node.state]
                    vm.last_state_change = int(time.time())
                    log.debug("VM: %s on %s. Changed from %s to %s." % (vm.id, self.name, vm.status, self.VM_STATES[node.state]))
            else:
                continue
        pass



import os
import sys
import time
import string
import json
import shutil
import logging
import nimbus_xml
import subprocess
import cluster_tools
import cloud_init_util
import cloudscheduler.config as config
import cloudscheduler.utilities as utilities
import datetime as dt
from cloudscheduler.job_management import _attr_list_to_dict
log = utilities.get_cloudscheduler_logger()
try:
    import boto.ec2
    import boto
except ImportError:
    log.error("To use EC2-style clouds, you need to have boto " \
            "installed. You can install it from your package manager, " \
            "or get it from http://code.google.com/p/boto/")
    
from httplib import BadStatusLine
from subprocess import Popen
from urlparse import urlparse
from cStringIO import StringIO
import gzip


class EC2Cluster(cluster_tools.ICluster):

    VM_STATES = {
            "running" : "Running",
            "pending" : "Starting",
            "shutting-down" : "Shutdown",
            "terminated" : "Shutdown",
            "error" : "Error",
    }

    ERROR = 1
    DEFAULT_INSTANCE_TYPE = config.default_VMInstanceType if config.default_VMInstanceType else "m1.small"
    DEFAULT_INSTANCE_TYPE_LIST = _attr_list_to_dict(config.default_VMInstanceTypeList)

    def _get_connection(self):
        """
            _get_connection - get a boto connection object to this cluster

            returns a boto connection object, or none in the case of an error
        """

        connection = None
        if len(self.regions) > 0:
            region_name = self.regions[0]
        else:
            region_name = self.name

        if self.cloud_type == "AmazonEC2":
            try:
                region = boto.ec2.regioninfo.RegionInfo(name=region_name,
                                                 endpoint=self.network_address)
                connection = boto.connect_ec2(
                                   aws_access_key_id=self.access_key_id,
                                   aws_secret_access_key=self.secret_access_key,
                                   region=region
                                   )
                log.verbose("Created a connection to Amazon EC2")

            except boto.exception.EC2ResponseError, e:
                log.error("Couldn't connect to Amazon EC2 because: %s" %
                                                                e.error_message)

        elif self.cloud_type == "Eucalyptus":
            try:
                region = boto.ec2.regioninfo.RegionInfo(name=region_name,
                                                 endpoint=self.network_address)
                connection = boto.connect_ec2(
                                   aws_access_key_id=self.access_key_id,
                                   aws_secret_access_key=self.secret_access_key,
                                   is_secure=self.secure_connection,
                                   region=region,
                                   port=8773,
                                   path="/services/Eucalyptus",
                                   )
                log.verbose("Created a connection to Eucalyptus (%s)" % self.name)

            except boto.exception.EC2ResponseError, e:
                log.error("Couldn't connect to Eucalyptus EC2 because: %s" %
                                                               e.error_message)

        elif self.cloud_type == "OpenNebula":

            log.error("OpenNebula support isn't ready yet.")
            raise NotImplementedError

        elif self.cloud_type == "OpenStack":
            try:
                region = boto.ec2.regioninfo.RegionInfo(name=region_name,
                                                 endpoint=self.network_address)
                connection = boto.connect_ec2(
                                   aws_access_key_id=self.access_key_id,
                                   aws_secret_access_key=self.secret_access_key,
                                   is_secure=self.secure_connection,
                                   region=region,
                                   port=8773,
                                   path="/services/Cloud",
                                   validate_certs=False
                                   )
                log.verbose("Created a connection to OpenStack (%s)" % self.name)

            except boto.exception.EC2ResponseError, e:
                log.error("Couldn't connect to OpenStack because: %s" %
                            e.error_message)
        else:
            log.error("EC2Cluster don't know how to handle a %s cluster." %
                                                               self.cloud_type)

        return connection

    def __init__(self, name="Dummy Cluster", host="localhost", cloud_type="Dummy",
                 memory=[], max_vm_mem= -1, networks=[], vm_slots=0,
                 cpu_cores=0, storage=0, access_key_id=None, secret_access_key=None,
                 security_group=None, hypervisor='xen', key_name=None, 
                 boot_timeout=None, secure_connection="", regions=[], vm_domain_name="",
                  reverse_dns_lookup=False,placement_zone=None, enabled=True, priority=0):

        # Call super class's init
        cluster_tools.ICluster.__init__(self,name=name, host=host, cloud_type=cloud_type,
                         memory=memory, max_vm_mem=max_vm_mem, networks=networks,
                         vm_slots=vm_slots, cpu_cores=cpu_cores,
                         storage=storage, hypervisor=hypervisor, boot_timeout=boot_timeout, enabled=enabled, priority=priority)

        if not security_group:
            security_group = ["default"]
        self.security_groups = security_group

        if not access_key_id or not secret_access_key:
            log.error("Cannot connect to cluster %s "
                      "because you haven't specified an access_key_id or "
                      "a secret_access_key" % self.name)

        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.key_name = key_name
        self.secure_connection = secure_connection in ['True', 'true', 'TRUE']
        self.total_cpu_cores = -1
        self.regions = regions
        self.vm_domain_name = vm_domain_name if vm_domain_name != None else ""
        self.reverse_dns_lookup = reverse_dns_lookup in ['True', 'true', 'TRUE']
        self.placement_zone = placement_zone

    def vm_create(self, vm_name, vm_type, vm_user, vm_networkassoc,
                  vm_image, vm_mem, vm_cores, vm_storage, customization=None,
                  pre_customization=None, vm_keepalive=0, instance_type="", 
                  maximum_price=0, job_per_core=False, securitygroup=[],
                  key_name="",use_cloud_init=False, extra_userdata=[]):
        """Attempt to boot a new VM on the cluster."""

        use_cloud_init = use_cloud_init or config.use_cloud_init
        log.verbose("Trying to boot %s on %s" % (vm_type, self.network_address))
        if len(securitygroup) != 0:
            sec_group = []
            for group in securitygroup:
                if group in self.security_groups:
                    sec_group.append(group)
            if len(sec_group) == 0:
                log.debug("No matching security groups - trying default config")
                sec_group = self.security_groups
                #sec_group.append("default") - don't just append default use what is in cloud_resources.conf for this cloud
        else:
            sec_group = self.security_groups

        try:
            if self.name in vm_image.keys():
                vm_ami = vm_image[self.name]
            else:
                vm_ami = vm_image[self.network_address]
        except:
            log.debug("No AMI for %s, trying default" % self.network_address)
            #try:
            #    vm_ami = vm_image["default"]
            #except:
                #log.debug("No given default - trying global defaults")
            try:
                vm_default_ami = _attr_list_to_dict(config.default_VMAMI)
                if self.name in vm_default_ami.keys():
                    vm_ami = vm_default_ami[self.name]
                else:
                    vm_ami = vm_default_ami[self.network_address]
            except:
                try:
                    vm_ami = vm_default_ami["default"]
                except:
                    log.exception("Can't find a suitable AMI")
                    self.failed_image_set.add(vm_ami)
                    return

        try:
            if self.name in instance_type.keys():
                i_type = instance_type[self.name]
            elif self.network_address in instance_type.keys():
                i_type = instance_type[self.network_address]
            else:
                i_type = instance_type["default"]
        except:
            log.debug("No instance type for %s, trying default" % self.network_address)
            #try:
            #    i_type = instance_type["default"]
            #except:
            #    if isinstance(instance_type, str):
            #        i_type = instance_type
            #    else:
            try:
                if self.name in self.DEFAULT_INSTANCE_TYPE_LIST.keys():
                    i_type = self.DEFAULT_INSTANCE_TYPE_LIST[self.name]
                else:
                    i_type = self.DEFAULT_INSTANCE_TYPE_LIST[self.network_address]
            except:
                log.debug("No default instance type found for %s, trying single default" % self.network_address)
                i_type = self.DEFAULT_INSTANCE_TYPE
        instance_type = i_type

        if key_name == None:
            key_name = self.key_name
        if customization:
            if not use_cloud_init:
                user_data = nimbus_xml.ws_optional(customization)
            else:
                user_data = cloud_init_util.build_write_files_cloud_init(customization)
        else:
            user_data = ""

        if pre_customization:
            if not use_cloud_init:
                for item in pre_customization:
                    user_data = '\n'.join([item, user_data])
            else:
                user_data = cloud_init_util.inject_customizations(pre_customization, user_data)
        elif use_cloud_init:
            user_data = cloud_init_util.inject_customizations([], user_data)[0]
        if len(extra_userdata) > 0:
            # need to use the multi-mime type functions
            user_data = cloud_init_util.build_multi_mime_message([(user_data, 'cloud-config', 'cloud_conf.yaml')], extra_userdata)

        if "AmazonEC2" == self.cloud_type and vm_networkassoc != "public":
            log.debug("You requested '%s' networking, but EC2 only supports 'public'" % vm_networkassoc)
            addressing_type = "public"
        else:
            addressing_type = vm_networkassoc

        try:
            connection = self._get_connection()
            image = None

            if not "Eucalyptus" == self.cloud_type:
                image = connection.get_image(vm_ami)

            else:   #HACK: for some reason Eucalyptus won't respond properly to
                    #      get_image("whateverimg"). Use a linear search until
                    #      this is fixed
                    # This is Eucalyptus bug #495670
                    # https://bugs.launchpad.net/eucalyptus/+bug/495670
                images = connection.get_all_images()
                for potential_match in images:
                    if potential_match.id == vm_ami:
                        image = potential_match
                        break

            # Compress the user data to try and get under the amazon limit
            udbuf = StringIO()
            udf = gzip.GzipFile(mode='wb', fileobj=udbuf)
            try:
                udf.write(user_data)
            finally:
                udf.close()
            user_data = udbuf.getvalue() 

            if image:
                if maximum_price is 0 or self.cloud_type == "OpenStack": # don't request a spot instance
                    try:
                        reservation = image.run(1,1, key_name=key_name,
                                                addressing_type=addressing_type,
                                                user_data=user_data,
                                                placement=self.placement_zone,
                                                security_groups=sec_group,
                                                instance_type=instance_type)
                        instance_id = reservation.instances[0].id
                        log.debug("Booted VM %s" % instance_id)
                    except boto.exception.EC2ResponseError, e:
                        log.exception("There was a problem creating an EC2 instance: %s" % e)
                        return self.ERROR
                    except Exception, e:
                        log.exception("There was an unexpected problem creating an EC2 instance: %s" % e)
                        return self.ERROR

                else: # get a spot instance of no more than maximum_price
                    try:
                        # try to get a better maximum_price
                        #log.info("vm_inst: %s  vm_type: %s  vm_image:%s"%(instance_type,vm_type,vm_image))
                        #spot_price = self.get_current_us_west_2_spot_price(instance_type,connection)
                        #log.info("Compare max_price %s with curr_price %s"%(maximum_price,spot_price))   
                        
                        #if maximum_price == 0 and isinstance(spot_price,float):
                        #    maximum_price = spot_price
                        
                        #if isinstance(spot_price,float) and maximum_price > spot_price:
                        #    maximum_price = spot_price
                            
                        #log.info("New max_price is %s",maximum_price)
                        reservation = connection.request_spot_instances(
                                                  maximum_price,
                                                  image.id,
                                                  key_name=key_name,
                                                  user_data=user_data,
                                                  placement=self.placement_zone,
                                                  addressing_type=addressing_type,
                                                  security_groups=sec_group,
                                                  instance_type=instance_type)
                        spot_id = str(reservation[0].id)
                        instance_id = ""
                        log.debug("Reserved instance %s at no more than %s" % (spot_id, maximum_price))
                    except AttributeError:
                        log.exception("Your version of boto doesn't seem to support "\
                                  "spot instances. You need at least 1.9")
                        return self.ERROR
                    except boto.exception.EC2ResponseError, e:
                        log.exception("There was a problem creating an EC2 spot instance: %s" % e)
                        return self.ERROR
                    except Exception, e:
                        log.exception("Problem an unexpected error creating an EC2 spot instance: %s" % e)
                        return self.ERROR


            else:
                log.error("Couldn't find image %s on %s" % (vm_image, self.name))
                self.failed_image_set.add(vm_ami)
                return self.ERROR
        
        except BadStatusLine, e:
            log.exception("Bad Status Line exception: %s" % e)
            return self.ERROR
        except Exception, e:
            log.exception("Problem creating EC2 instance on %s: %s" % (self.name, e))
            try:
                if e.errors and len(e.errors) > 0 and len(e.errors[0]) > 0 and e.errors[0][0] == "ImageNotFound":
                    self.failed_image_set.add(vm_ami)
            except:
                return self.ERROR
            return self.ERROR

        vm_mementry = self.find_mementry(vm_mem)
        if (vm_mementry < 0):
            #TODO: this is kind of pointless with EC2...
            log.debug("Cluster memory list has no sufficient memory " +\
                      "entries (Not supposed to happen). Returning error.")
            return self.ERROR
        log.verbose("vm_create - Memory entry found in given cluster: %d" %
                                                                    vm_mementry)
        new_vm = cluster_tools.VM(name = vm_name, id = instance_id, vmtype = vm_type, user = vm_user,
                    clusteraddr = self.network_address,
                    cloudtype = self.cloud_type, network = vm_networkassoc,
                    image= vm_ami, flavor=instance_type,
                    memory = vm_mem, mementry = vm_mementry,
                    cpucores = vm_cores, storage = vm_storage, 
                    keep_alive = vm_keepalive, job_per_core = job_per_core)

        try:
            new_vm.spot_id = spot_id
        except:
            log.verbose("No spot ID to add to VM %s" % instance_id)

        try:
            self.resource_checkout(new_vm)
        except:
            log.exception("Unexpected Error checking out resources when creating a VM. Programming error?")
            self.vm_destroy(new_vm, reason="Failed Resource checkout")
            return self.ERROR

        self.vms.append(new_vm)

        return 0


    def vm_poll(self, vm):
        """Query the cloud service for information regarding a VM."""
        try:
            log.verbose("Polling vm with instance id %s" % vm.id)
            connection = self._get_connection()

            if vm.spot_id:
                try:
                    spot_reservation = connection.get_all_spot_instance_requests(vm.spot_id)[0]
                    if spot_reservation.instance_id == None:
                        log.debug("Spot reservation %s doesn't have a VM id yet." % vm.spot_id)
                        return vm.status
                    vm.id = str(spot_reservation.instance_id)
                except AttributeError:
                    log.exception("Problem getting spot VM info. Do you have boto 2.0+?")
                    return vm.status
                except:
                    log.exception("Problem getting information for spot vm %s" % vm.spot_id)
                    return vm.status

            instance = None
            try:
                reservations = connection.get_all_instances([vm.id])
                instance = reservations[0].instances[0]
            except IndexError:
                log.error("%s on %s doesn't seem to exist anymore, setting status to Error" % (vm.id, self.network_address))
                vm.status = self.VM_STATES['error']
                vm.last_state_change = int(time.time())
                return vm.status
            except boto.exception.EC2ResponseError, e:
                log.exception("Unexpected error polling %s: %s" % (vm.id, e))
                if e.status == 400 and e.error_code == 'InstanceNotFound':
                    vm.status = self.VM_STATES['error']
                elif e.status == 404:
                    vm.status = self.VM_STATES['error']
                return vm.status
            except Exception, e:
                log.exception("Unexpected exception polling vm: %s on: %s: %s" % (vm.id, self.name, e))
                return vm.status

        except boto.exception.EC2ResponseError, e:
            log.error("Couldn't update status because: %s" % e.error_message)
            return vm.status

        if not instance:
            return vm.status
        with self.vms_lock:
            if instance and vm.status != self.VM_STATES.get(instance.state, "Starting"):

                vm.last_state_change = int(time.time())
                log.debug("VM: %s on %s. Changed from %s to %s." % (vm.id, self.name, vm.status, self.VM_STATES.get(instance.state, "Starting")))
            vm.status = self.VM_STATES.get(instance.state, "Starting")
            if self.reverse_dns_lookup:
                # run a dig -x on the ip address
                dig_cmd = ['dig', '-x', instance.ip_address]
                (_, dig_out, _) = self.vm_execwait(dig_cmd, env=vm.get_env())
                # extract the hostname from dig -x output
                vm.hostname = self._extract_host_from_dig(dig_out)
            elif self.cloud_type == "OpenStack":
                if len(instance.public_dns_name) > 0:
                    vm.hostname = ''.join([instance.public_dns_name, self.vm_domain_name])
                else:
                    vm.hostname = ''.join([instance.private_dns_name, self.vm_domain_name])
            else:
                if len(instance.public_dns_name) > 0:
                    vm.hostname = instance.public_dns_name
                else:
                    vm.hostname = instance.private_dns_name
            if len(instance.public_dns_name) > 0 and len(instance.private_dns_name) > 0:
                vm.hostname = instance.public_dns_name
                vm.alt_hostname = instance.private_dns_name
                if self.cloud_type == "OpenStack":
                    vm.hostname = ''.join([vm.hostname, self.vm_domain_name])
                    vm.alt_hostname = ''.join([vm.alt_hostname, self.vm_domain_name])
            vm.lastpoll = int(time.time())
        return vm.status


    def vm_destroy(self, vm, return_resources=True, reason=""):
        """
        Shutdown, destroy and return resources of a VM to it's cluster

        Parameters:
        vm -- vm to shutdown and destroy
        return_resources -- if set to false, do not return resources from VM to cluster
        """
        log.info("Destroying VM: %s Name: %s Reason: %s" % (vm.id, vm.hostname, reason))

        try:
            connection = self._get_connection()

            if vm.spot_id:
                connection.cancel_spot_instance_requests([vm.spot_id])

            if vm.id:
                reservations = connection.get_all_instances([vm.id])
                instance = reservations[0].instances[0]
                instance.terminate()

        except IndexError:
            log.warning("%s already seem to be gone... removing anyway." % vm.id)
        except boto.exception.EC2ResponseError, e:
            returnError = True
            log.exception("Couldn't connect to cloud to destroy VM: %s !" % vm.id)
            if e.status == 400 and e.error_code == 'InstanceNotFound':
                log.exception("VM %s no longer exists... removing from system" % vm.id)
                returnError = False
            if e.status == 404:
                log.exception("VM %s not found... removing from system" % vm.id)
                returnError = False
            if returnError:
                return self.ERROR
        except:
            log.exception("Unexpected error destroying VM: %s!" % vm.id)

        # Delete references to this VM
        if return_resources and vm.return_resources:
            self.resource_return(vm)
        with self.vms_lock:
            try:
                self.vms.remove(vm)
            except Exception as e:
                log.error("Unable to remove VM %s on %s: %s" % (vm.id, self.name, e))

        return 0

    def get_vm(self, vm_id):
        """Get VM object with id value. Override to also check spot_id"""
        for vm in self.vms:
            if vm_id == vm.id:
                return vm
        for vm in self.vms:
            if vm_id == vm.spot_id:
                return vm
        return None

    def _extract_host_from_dig(self, dig_out):
        at_answer_line = False
        hostname = ""
        for line in dig_out.split('\n'):
            parts = line.split()
            if at_answer_line:
                hostname = parts[-1][:-1]
                break
            elif 'ANSWER' in parts and 'SECTION:' in parts:
                at_answer_line = True
                continue
        return hostname

    def vm_execwait(self, cmd, env=None):
        """As above, a function to encapsulate command execution via Popen.
        vm_execwait executes the given cmd list, waits for the process to finish,
        and returns the return code of the process. STDOUT and STDERR are stored
        in given parameters.
        Parameters:
        (cmd as above)
        Returns:
        ret - The return value of the executed command
        out - The STDOUT of the executed command
        err - The STDERR of the executed command
        The return of this function is a 3-tuple
        """
        out = ""
        err = ""
        try:
            sp = Popen(cmd, shell=False,
                       stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
            if not utilities.check_popen_timeout(sp):
                (out, err) = sp.communicate(input=None)
            else:
                log.warning("Process %s timed out! cmd was %" % (sp.pid, " ".join(cmd)))
            return (sp.returncode, out, err)
        except OSError, e:
            try:
                log.error("Problem running %s, got errno %d \"%s\"" % (string.join(cmd, " "), e.errno, e.strerror))
            except:
                log.error("Problem running command, OSError.")
            return (-1, "", "")
        except:
            try:
                log.error("Problem running %s, unexpected error: %s" % (string.join(cmd, " "), err))
            except:
                log.error("Problem running command, unexpected error.")
            return (-1, "", "")

        
    """ These methods relate to inquiring on EC2 spot pricing methods """
    
    """ img_type examples 't1.micro','m3.medium','c3.2xlarge','m3.large','cc2.8xlarge','m1.medium' """
    def get_current_us_west_2_spot_price(self,img_type,connection):
        result=self._get_current_spot_price(["Linux/UNIX"], [img_type], [self.regions], [], "none", "json",connection)
        lowest_price = result[0]['price']
        for idx,price in enumerate(result):
            if lowest_price>result[idx]['price']:
                lowest_price=result[idx]['price']
        return lowest_price*1.1

    """ Support Methods for spot pricing methods described above - methods can be used to answer spot pricing questions"""
    
    def _get_current_spot_price(self,image_type,instance_type,regions,zones,filter_type,response_type,connection):
        """API function
        image_type $string see help -h 
        instance_type $string see help -h 
        regions @array of strings see help -h 
        zones @array of strings see help -h 
        filter_type $string see help -h 
        response_type $string see help -h 
        """
        args = []
        args.append( {"ImageTypes":image_type} )  
        args.append({"InstanceTypes":  instance_type })
        args.append({"Regions": regions})
        args.append({"ZONES": zones })
        json_req = self._create_request('current_spot_price',filter_type,response_type,{},args)
        json_req['start_time'] = dt.datetime.now().isoformat()
        json_req['end_time'] = dt.datetime.now().isoformat()
        json_response_data = self.__get_current_spot_price(json_req,connection)
        json_response_data = self.filter_response(json_response_data, json_req['filter_type'])
        return json_response_data
    
    def _create_request(self,req_type,filter_type,response_type,other_args,args):
        if req_type == 'current_spot_price' or req_type == 'list_all_spot_price' or req_type == 'periodic_spot_price':
            msg = {}
            msg['filter_type']=filter_type
            msg['response_type']=response_type
            msg['request']=req_type
            msg['other_args']=other_args
            msg['request_args']=args
            return msg
        
    def __get_current_spot_price(self,json_req,connection):
        req_args_array = json_req['request_args']
        regions=[]
        inst_types=[]
        image_types=[]
        zones = []
        
        json_response_data = []
        
        for obj in req_args_array:
            for obj_key in obj.keys():
                if obj_key == 'Regions':
                    regions = obj[obj_key];
                if obj_key == 'ZONES':
                    zones = obj[obj_key];
                if obj_key == 'InstanceTypes':
                    inst_types = obj[obj_key];
                if obj_key == 'ImageTypes':
                    image_types = obj[obj_key];
        
        for region in regions:
            #check zones 
            #conn = boto.ec2.connect_to_region(self.access_key_id,self.secret_access_key,region)
            region = boto.ec2.regioninfo.RegionInfo(name=self.regions,
                                                 endpoint='ec2.us-west-2.amazonaws.com')
            conn = boto.connect_ec2(
                                   aws_access_key_id=self.access_key_id,
                                   aws_secret_access_key=self.secret_access_key,
                                   region=region)
            
            #conn = connection
            if len(zones) > 0:
                #for each zone do calls for each image and eash instance type   
                for zone in zones:
                    for inst in inst_types:
                        for img in image_types:
                            
                            start_time = json_req['start_time']
                            end_time = json_req['end_time']
                            
                            instance_type=inst
                            product_description = img
                            availability_zone = zone
                            dry_run = False
                            max_results = 100
                            next_token = None
                            filters = None
                            price = conn.get_spot_price_history(start_time, end_time, instance_type, product_description, availability_zone, dry_run, max_results, next_token, filters)
                            #price_obj = price.pop()
                            for price_obj in price:
                                #pprint("%s : %s "%(price_obj,price_obj.price))
                                json_data = {}
                                json_data['region']=region
                                json_data['instance_type']=inst
                                json_data['image_type']=img
                                json_data['zone']=zone
                                json_data['price']=price_obj.price
                                json_data['timestamp']=price_obj.timestamp
                                json_response_data.append(json_data)
                                #price_obj = price.pop()
                            
            else:
                #for each region get their zones and do calles for each instance type and image type
                #conn = boto.ec2.connect_to_region(region)
                reg_zones = conn.get_all_zones()
                for zone in reg_zones:
                    for inst in inst_types:
                        for img in image_types:
                            start_time = json_req['start_time']
                            end_time = json_req['end_time']
                            instance_type=inst
                            product_description = img
                            availability_zone = zone.name
                            dry_run = False
                            max_results = 100
                            next_token = None
                            filters = None
                            price = conn.get_spot_price_history(start_time, end_time, instance_type, product_description, availability_zone, dry_run, max_results, next_token, filters)
                            
                            for price_obj in price:
                                json_data = {}
                                json_data['region']=region
                                json_data['instance_type']=inst
                                json_data['image_type']=img
                                json_data['zone']=zone.name
                                json_data['price']=price_obj.price
                                json_data['timestamp']=price_obj.timestamp
                                json_response_data.append(json_data)
        
        return json_response_data
    
    def filter_response(self,json_resp, filter_type):
        if filter_type == "none":
            return json_resp
        elif filter_type == "highest":
            highest=0.0
            highest_idx=-1;
            remove_idx=[]
            
            for idx,entry in enumerate(json_resp):
                if entry['price'] <= highest:
                    remove_idx.append(idx)
                else:
                    remove_idx.append(idx)
                    highest_idx=idx
                    highest=entry['price']
            
            if highest_idx != -1:
                del remove_idx[highest_idx]
            while len(remove_idx)>0:
                idx = remove_idx.pop()
                del json_resp[idx]
                
            return json_resp
        elif filter_type == "average":
            total=0.0
            idx=0
            remove_idx=[]
           
            for idx,entry in enumerate(json_resp):
                total=total+entry['price']
                remove_idx.append(idx)
            average=total/(1+idx)
            while len(remove_idx)>0:
                idx = remove_idx.pop()
                del json_resp[idx]
            json_str = '{"zone":"","image_type":"","timestamp":"","region":"","instance_type":"","price":%s}'%average
            
            json_obj = json.loads(json_str)
            json_resp.append(json_obj)   
                
            return json_resp


import os
import time
import uuid
import threading
import nimbus_xml
import ConfigParser
import cluster_tools
import cloud_init_util
import base64
import cloudscheduler.config as config
import cloudscheduler.utilities as utilities
from cloudscheduler.job_management import _attr_list_to_dict
try:
    import httplib2
    from oauth2client.client import flow_from_clientsecrets
    from oauth2client.file import Storage
    from oauth2client.tools import run
    from oauth2client.tools import run_flow
    from apiclient.discovery import build
except:
    pass

log = utilities.get_cloudscheduler_logger()


class GoogleComputeEngineCluster(cluster_tools.ICluster):
    VM_STATES = {
            "RUNNING" : "Running",
            "Starting" : "Starting",
            "shutting-down" : "Shutdown",
            "terminated" : "Shutdown",
            "PROVISIONING" : "Provisioning",
            "error" : "Error",
    }
    
    GCE_SCOPE = 'https://www.googleapis.com/auth/compute'
    
    API_VERSION = 'v1'
    GCE_URL = 'https://www.googleapis.com/compute/%s/projects/' % (API_VERSION)

    DEFAULT_ZONE = 'us-central1-b' # will need to be option in job
    DEFAULT_MACHINE_TYPE = 'n1-standard-1'  # option specified in job config
    DEFAULT_INSTANCE_TYPE_LIST = _attr_list_to_dict(config.default_VMInstanceTypeList)
    DEFAULT_IMAGE = 'cloudscheduler-centos-9'
    DEFAULT_ROOT_PD_NAME = 'hepgc-uvic-root-pd'  

    DEFAULT_NETWORK = 'default' # job option setup
    DEFAULT_SERVICE_EMAIL = 'default' 
    DEFAULT_SCOPES = ['https://www.googleapis.com/auth/devstorage.full_control',
                  'https://www.googleapis.com/auth/compute']

    def __init__(self, name="Dummy Cluster", host="localhost",
                 cloud_type="Dummy", memory=[], max_vm_mem= -1, networks=[],
                 vm_slots=0, cpu_cores=0, storage=0, hypervisor='xen', boot_timeout=None,
                 auth_dat_file=None, secret_file=None, security_group=None, project_id=None,enabled=True, priority = 0,
                 total_cpu_cores=-1):
        log.debug("Init GCE cores %s, storage %s"%(cpu_cores,storage))
        self.gce_hostname_prefix = 'gce-cs-vm'
        self.security_group = security_group
        self.auth_dat_file_path = auth_dat_file
        self.secret_file_path = secret_file
        self.project_id = project_id
        self.total_cpu_cores = total_cpu_cores
        if not project_id:
            return None
        
        
        # Perform OAuth 2.0 authorization.
        flow = flow_from_clientsecrets(self.secret_file_path, scope=self.GCE_SCOPE)
        auth_storage = Storage(self.auth_dat_file_path)
        credentials = auth_storage.get()
      
        if credentials is None or credentials.invalid:
            credentials = run_flow(flow, auth_storage)
        http = httplib2.Http()
        self.auth_http = credentials.authorize(http)

        #if not security_group:
        #    security_group = ["default"]
        #self.security_groups = security_group
        
        # Build service object
        self.gce_service = build('compute', self.API_VERSION)
        self.project_url = '%s%s' % (self.GCE_URL, self.project_id)
        # Call super class's init
        
        cluster_tools.ICluster.__init__(self,name=name, host=host, cloud_type=cloud_type,
                         memory=memory, max_vm_mem=max_vm_mem, networks=networks,
                         vm_slots=vm_slots, cpu_cores=cpu_cores,
                         storage=storage, hypervisor=hypervisor, boot_timeout=boot_timeout,enabled=enabled, priority=priority)

    def vm_create(self, vm_name, vm_type, vm_user, vm_networkassoc,
                  vm_image, vm_mem, vm_cores, vm_storage, customization=None,
                  vm_keepalive=0, instance_type="", maximum_price=0,
                  job_per_core=False, securitygroup=[],pre_customization=None,use_cloud_init=False, extra_userdata=[]):
        
        try:
            if self.network_address in vm_image.keys():
                vm_ami = vm_image[self.network_address]
            elif self.name in vm_image.keys():
                vm_ami = vm_image[self.name]
        except:
            log.debug("No AMI for %s, trying default" % self.network_address)
            try:
                vm_ami = vm_image["default"]
            except:
                log.exception("Can't find a suitable AMI")
                return
        # Construct URLs
       
        try:
            if self.name in instance_type.keys():
                i_type = instance_type[self.name]
            else:
                i_type = instance_type[self.network_address]
        except:
            log.debug("No instance type for %s, trying default" % self.network_address)
            try:
                if self.name in self.DEFAULT_INSTANCE_TYPE_LIST.keys():
                    i_type = self.DEFAULT_INSTANCE_TYPE_LIST[self.name]
                else:
                    i_type = self.DEFAULT_INSTANCE_TYPE_LIST[self.network_address]
            except:
                log.debug("No default instance type found for %s, trying single default" % self.network_address)
                i_type = self.DEFAULT_MACHINE_TYPE
        vm_instance_type = i_type
        
        if vm_image:
            vm_image_name = vm_ami
        else:
            vm_image_name = self.DEFAULT_IMAGE

        #this should replace disk_url when cloud-init supports GCE in CERNVM3
        image_url = '%s%s/global/images/%s' % (
              self.GCE_URL, self.project_id, vm_image_name)
        
        #Ensures the VM's Root Disks are Unique
        self.DEFAULT_ROOT_PD_NAME = '%s-%s'%('hepgc-uvic-root-pd',self.generate_next_instance_name())
        
        #temporary variable for disk_url
        #https://www.googleapis.com/compute/v1/projects/atlasgce/zones/us-central1-b/disks/hepgc-uvic-root-pd
        disk_url = '%s%s/zones/%s/disks/%s'%(self.GCE_URL,self.project_id,self.DEFAULT_ZONE,self.DEFAULT_ROOT_PD_NAME)
        
        machine_type_url = '%s/zones/%s/machineTypes/%s' % (
              self.project_url, self.DEFAULT_ZONE, vm_instance_type)
        #zone_url = '%s/zones/%s' % (self.project_url, self.DEFAULT_ZONE)
        network_url = '%s/global/networks/%s' % (self.project_url, self.DEFAULT_NETWORK)
        
        # Construct the request body
        disk = {
               'name': self.DEFAULT_ROOT_PD_NAME,
               'sourceSnapshot':'https://www.googleapis.com/compute/v1/projects/atlasgce/global/snapshots/%s'%vm_image_name,
               'sizeGb':vm_storage
        }

        # Create the root pd
        try:
            request = self.gce_service.disks().insert(project=self.project_id, body=disk, zone=self.DEFAULT_ZONE)
            response = request.execute(http=self.auth_http)
            response = self._blocking_call(self.gce_service, self.auth_http, response)
        except:
            log.exception('Error Trying to create disk, one already exists ... returning ')
            return
        use_cloud_init = use_cloud_init or config.use_cloud_init
        if customization:
            if not use_cloud_init:
                user_data = nimbus_xml.ws_optional(customization)
            else:
                user_data = cloud_init_util.build_write_files_cloud_init(customization)
        else:
            user_data = ""
        
        
        if pre_customization:
            if not use_cloud_init:
                for item in pre_customization:
                    user_data = '\n'.join([item, user_data])
            else:
                user_data = cloud_init_util.inject_customizations(pre_customization, user_data)
        elif use_cloud_init:
            user_data = cloud_init_util.inject_customizations([], user_data)[0]
        if len(extra_userdata) > 0:
            # need to use the multi-mime type functions
            user_data = cloud_init_util.build_multi_mime_message([(user_data, 'cloud-config')], extra_userdata)     
        next_instance_name = self.generate_next_instance_name()
        
        instance = {
          'name': next_instance_name,
          'machineType': machine_type_url,
          #'disks': [{
          #      'autoDelete': 'true',
          #      'boot': 'true',
          #      'type': 'PERSISTENT',
          #      'initializeParams' : {
          #              'diskname': self.DEFAULT_ROOT_PD_NAME,
          #              'sourceImage': image_url
          #              }
          #      }],
          'disks': [{
                'source':disk_url,
                'boot': 'true',
                'autoDelete':'true',
                'type': 'PERSISTENT',
                }],
          #'image': image_url,
          'networkInterfaces': [{
            'accessConfigs': [{
              'type': 'ONE_TO_ONE_NAT',
              'name': 'External NAT'
             }],
            'network': network_url
          }],
          'serviceAccounts': [{
               'email': self.DEFAULT_SERVICE_EMAIL,
               'scopes': self.DEFAULT_SCOPES
          }],
          'metadata': {
              'items': [
                {
                  'key': 'user-data',
                  'value': user_data ,
                },
                
            ]
          }
        }
        
        
        # Create the instance
        response = None
        request = self.gce_service.instances().insert(
             project=self.project_id, body=instance, zone=self.DEFAULT_ZONE)
        try:
            response = request.execute(http=self.auth_http)
            response = self._blocking_call(self.gce_service, self.auth_http, response)
            #log.info('Created VM ')
        except Exception, e:
            log.error("Error creating VM on gce: %s" % e)
            pass

        if response and 'targetId' in response:
            target_id = response['targetId']
        elif response:
            return
        else:
            #print 'no response'
            return
        vm_mementry = self.find_mementry(vm_mem)
        if (vm_mementry < 0):
            #TODO: this is kind of pointless with EC2..., but the resource code depends on it
            log.debug("Cluster memory list has no sufficient memory " +\
                      "entries (Not supposed to happen). Returning error.")
            return self.ERROR
        new_vm = cluster_tools.VM(name = next_instance_name, vmtype = vm_type, user = vm_user,
                    clusteraddr = self.network_address, id = target_id,
                    cloudtype = self.cloud_type, network = vm_networkassoc,
                    hostname = self.construct_hostname(next_instance_name),
                    cpuarch = "x86", image= vm_image, mementry = vm_mementry, flavor=vm_instance_type,
                    memory = vm_mem, cpucores = vm_cores, storage = vm_storage, 
                    keep_alive = vm_keepalive, job_per_core = job_per_core)
    
        try:
            self.resource_checkout(new_vm)
        except:
            log.exception("Unexpected Error checking out resources when creating a VM. Programming error?")
            self.vm_destroy(new_vm, reason="Failed Resource checkout")
            return self.ERROR
    
        self.vms.append(new_vm)
        #log.info("added a new vm %s"%new_vm)
        return 0


    def vm_destroy(self, vm, return_resources=True, reason=""):
        # Delete an Instance
        log.info("googlecluster::destroy vm::%s"%vm.name)
        request = self.gce_service.instances().delete(
            project=self.project_id, instance=vm.name, zone=self.DEFAULT_ZONE)
        try:
            response = request.execute(http=self.auth_http)
            #log.info("Destroy vm, response execute  %s ret res %s"%(vm.name,return_resources))
            response = self._blocking_call(self.gce_service, self.auth_http, response)
            #log.info("Destroy vm, response waiting  %s"%vm.name)
        except:
            log.error("Failure while destroying VM %s. return leaving with removing resource from cloud sched" % (vm.name))
            return
        #log.info("Destroy VM %s, check response %s"%(vm.name,response))
        try:
            if response and response['status'] == 'DONE':
                # Delete references to this VM
                if return_resources and vm.return_resources:
                    self.resource_return(vm)
                with self.vms_lock:
                    self.vms.remove(vm)
                return 0
            else:
                log.debug("Error Destroying GCE VM: %s" % (vm.name))
                return 1
        except:
            log.exception("Error removing vm, possibly already removed")
            return 1

    def vm_poll(self, vm):
        #filter_str = ''.join(["id eq ", vm.id])
        #request = self.gce_service.instances().list(project=self.project_id, filter=filter_str)
        request = self.gce_service.instances().list(project=self.project_id, filter=None, zone=self.DEFAULT_ZONE)
        try:
            response = request.execute(http=self.auth_http)
        except Exception as e:
            log.error("Problem polling gce vm %s error %s will retry later." % (vm.id, e))
            return

        if response and 'items' in response:
            instances = response['items']
            
            for instance in instances:
                
                if 'id' in instance and instance['id'] == vm.id:
                    log.info("googlecluster::state::%s::inst %s::vm %s"%(vm.name,instance['status'],vm.status))
                    if instance and hasattr(vm, 'status') and vm.status != self.VM_STATES.get(instance['status'], "Starting"):
                        vm.last_state_change = int(time.time())
                        
                
                    vm.status = self.VM_STATES[instance['status']]
                    vm.ipaddress = instance['networkInterfaces'][0]['accessConfigs'][0]['natIP']
                
        else:
            pass


    def _blocking_call(self, gce_service, auth_http, response):
        """Blocks until the operation status is done for the given operation."""
        if 'status' in response:
            status = response['status']
            
        while status != 'DONE' and response:
            if 'name' in response:
                operation_id = response['name']
            else:
                break
            if 'zone' in response:
                zone_name = response['zone'].split('/')[-1]
                request = gce_service.zoneOperations().get(project=self.project_id, operation=operation_id, zone=zone_name)
            else:
                request = gce_service.globalOperations().get(project=self.project_id, operation=operation_id)
            try:
                response = request.execute(http=auth_http)
            except:
                pass
            
            if response and 'status' in response:
                status = response['status']

            time.sleep(1)
        return response
    
    def generate_next_instance_name(self):
        potential_name = ''.join([self.gce_hostname_prefix, str(uuid.uuid4())])
        potential_name = potential_name[0:15]
        collision = False
        for vm in self.vms:
            if potential_name == vm.name:
                collision = True
                break
        if collision:
            potential_name = None
        return potential_name
    def construct_hostname(self, instance_name):
        return ''.join([instance_name, '.c.', self.project_id, '.internal'])
            



#!/usr/bin/env python
# vim: set expandtab ts=4 sw=4:

# Copyright (C) 2009 University of Victoria
# You may distribute under the terms of either the GNU General Public
# License or the Apache v2 License, as specified in the README file.


##
## This file contains the VM class, ICluster interface, as well as the
## implementations of this interface.

from __future__ import with_statement

import os
import re
import sys
import time
import string
import shutil
import logging
import datetime
import tempfile
import subprocess
import threading

from subprocess import Popen
from urlparse import urlparse

import nimbus_xml
import config
import cloudscheduler.utilities as utilities
from cloudscheduler.utilities import get_cert_expiry_time

log = utilities.get_cloudscheduler_logger()


class VM:
    """
    A class for storing created VM information. Used to populate Cluster classes
    'vms' lists.

    Instance Variables

    The global VM states are:
       Starting - The VM is being created in the cloud
       Running  - The VM is running somewhere on the cloud (fully functional)
       Error    - The VM has been corrupted or is in the process of being destroyed

    States are defined in each Cluster subclass, in which a VM_STATES dictionary
    maps specific cloud software state to these global states.
    """

    def __init__(self, name="", id="", vmtype="", user="",
            hostname="", ipaddress="", clusteraddr="", clusterport="",
            cloudtype="", network="public", cpuarch="x86",
            image="", memory=0, mementry=0, flavor="",
            cpucores=0, storage=0, keep_alive=0, spot_id="",
            proxy_file=None, myproxy_creds_name=None, myproxy_server=None, myproxy_server_port=None, 
            myproxy_renew_time="12", job_per_core=False):
        """
        Constructor

        name         - (str) The name of the vm (arbitrary)
        id           - (str) The id tag for the VM. Whatever is used to access the vm
                       by cloud software (Nimbus: epr file. OpenNebula: id number, etc.)
        vmtype       - (str) The condor VMType attribute for the VM
        user         - (str) The user who 'owns' this VM
        uservmtype   - (str) Aggregate type in form 'user:vmtype'
        hostname     - (str) The first part of hostname given to VM
        ipaddress    - (str) The IP Address of the VM
        condorname   - (str) The name of the VM as it's registered with Condor
        condoraddr   - (str) The Address of the VM as it's registered with Condor
        clusteraddr  - (str) The address of the cluster hosting the VM
        clusterport  - (str) The port of the cluster hosting the VM
        cloudtype    - (str) The cloud type of the VM (Nimbus, OpenNebula, etc)
        network      - (str) The network association the VM uses
        cpuarch      - (str) The required CPU architecture of the VM
        image        - (str) The location of the image from which the VM was created
        memory       - (int) The memory used by the VM
        mementry     - (int) The index of the entry in the host cluster's memory list
                       from which this VM is taking memory
        flavor       - (str) The flavor/instance_type of the VM
        proxy_file   - the proxy that was used to authenticate this VM's creation
        myproxy_creds_name - (str) The name of the credentials to retreive from the myproxy server
        myproxy_server - (str) The hostname of the myproxy server to retreive user creds from
        myproxy_server_port - (str) The port of the myproxy server to retreive user creds from
        errorcount   - (int) Number of Polling Errors VM has had
        force_retire - (bool) Flag to prevent a retiring VM from being turned back on
        return_resources - (bool) Flag to set if resources from this VM should be returned to cluster
        """
        self.name = name
        self.id = id
        self.vmtype = vmtype
        self.user = user
        self.uservmtype = ':'.join([user,vmtype])
        self.hostname = hostname
        self.alt_hostname = None
        self.ipaddress = ipaddress
        self.condorname = None
        self.condoraddr = None
        self.condormasteraddr = None
        self.clusteraddr = clusteraddr
        self.clusterport = clusterport
        self.cloudtype = cloudtype
        self.network = network
        self.image = image
        self.memory = memory
        self.mementry = mementry
        self.flavor = flavor
        self.cpucores = cpucores
        self.storage = storage
        self.errorcount = 0
        self.errorconnect = None
        self.lastpoll = None
        self.last_state_change = None
        self.initialize_time = int(time.time())
        self.startup_time = None
        self.keep_alive = keep_alive
        self.idle_start = None
        self.spot_id = spot_id
        self.proxy_file = proxy_file
        self.myproxy_creds_name = myproxy_creds_name
        self.myproxy_server = myproxy_server
        self.myproxy_server_port = myproxy_server_port
        self.myproxy_renew_time = myproxy_renew_time
        self.override_status = None
        self.job_per_core = job_per_core
        self.force_retire = False
        self.return_resources = True
        self.failed_retire = False
        self.job_run_times = utilities.JobRunTrackQueue('Run_Times')
        self.x509userproxy_expiry_time = None
        
        # Set a status variable on new creation
        self.status = "Starting"

        global log
        log = logging.getLogger("cloudscheduler")
        log.verbose("New VM Object - Name: %s, id: %s, host: %s, image: %s, memory: %d" \
          % (name, id, clusteraddr, image, memory))
        log.info("Created VM cluster address %s name %s"%(clusteraddr,name))

    def log(self):
        """Log the VM to the info level."""
        log.info("VM Name: %s, ID: %s, Type: %s, User: %s, Status: %s on %s" % (self.name, self.id, self.vmtype,  self.user, self.status, self.clusteraddr))
    def log_dbg(self):
        """Log the VM to the debug level."""
        log.debug("VM Name: %s, ID: %s, Type: %s, User: %s, Status: %s on %s" % (self.name, self.id, self.vmtype, self.user, self.status, self.clusteraddr))

    def get_vm_info(self):
        """Formatted VM information for use with cloud_status."""
        nameout = self.hostname
        if self.hostname == "":
            nameout = self.ipaddress
        idout = self.id
        if self.id == "":
            idout = self.spot_id
        output = "%-6s %-25s %-20s %-10s %-12s\n" % (idout, nameout, self.vmtype, self.user, self.status)
        if self.override_status != None:
            output = "%-6s %-25s %-20s %-10s %-12s\n" % (idout, nameout, self.vmtype, self.user, self.override_status)
        return output

    @staticmethod
    def get_vm_info_header():
        """Formatted header for use with cloud_status vm info output."""
        return "%-6s %-25s %-20s %-10s %-12s %-23s\n" % ("ID", "HOSTNAME", "VMTYPE", "USER", "STATUS", "CLUSTER")

    def get_vm_info_pretty(self):
        """Header + VM info formatted output."""
        output = self.get_vm_info_header()
        output += self.get_vm_info()
        return output

    def get_proxy_file(self):
        """Return the proxy file associated with the VM or None."""
        if hasattr(self, "proxy_file"):
            return self.proxy_file
        else:
            return None

    def get_myproxy_creds_name(self):
        """Return the MyProxy credentials name associated with the VM or None."""
        if hasattr(self, "myproxy_creds_name"):
            return self.myproxy_creds_name
        else:
            return None

    def get_myproxy_server(self):
        """Return the MyProxy server associated with the VM or None."""
        if hasattr(self, "myproxy_server"):
            return self.myproxy_server
        else:
            return None

    def get_myproxy_server_port(self):
        """Return the MyProxy server port associated with the VM or None."""
        if hasattr(self, "myproxy_server_port"):
            return self.myproxy_server_port
        else:
            return None

    def get_renew_time(self):
        """Return the MyProxy proxy renewal time associated with the VM."""
        return self.myproxy_renew_time

    def get_x509userproxy_expiry_time(self):
        """Use this method to get the expiry time of the VM's user proxy, if any.
        Note that lazy initialization is done;  the expiry time will be extracted
        from the user proxy the first time the method is called and then it will
        be cached in the instance variable.

        Returns the expiry time as a datetime.datetime instance (UTC), or None
        if there is no user proxy associated with this VM.
        """
        if (self.x509userproxy_expiry_time == None) and (self.get_proxy_file() != None):
            self.x509userproxy_expiry_time = get_cert_expiry_time(self.get_proxy_file())
        return self.x509userproxy_expiry_time

    def reset_x509userproxy_expiry_time(self):
        """Use this method to trigger an update of the proxy expiry time next
        time it is checked. For example, this must be called right after the
        proxy has been renewed. See get_x509userproxy_expiry_time for more info
        about how the proxy expiry time is cached in memory.
        """
        self.x509userproxy_expiry_time = None


    def is_proxy_expired(self):
        """Test if a VM's user proxy is expired.

        Returns True if the proxy is expired, False otherwise.
        """
        expiry_time = self.get_x509userproxy_expiry_time()
        if expiry_time == None:
            return False
        return expiry_time <= datetime.datetime.utcnow()


    def needs_proxy_renewal(self):
        """Test if a VM's user proxy needs to be refreshed, according
        the VM proxy refresh threshold found in the cloud scheduler configuration.

        Returns: True if the proxy needs to be refreshed, or 
                 False otherwise (or if the VM has no user proxy associated with it).
        """
        expiry_time = self.get_x509userproxy_expiry_time()
        if expiry_time == None:
            return False
        td = expiry_time - datetime.datetime.utcnow()
        td_in_seconds = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
        log.verbose("needs_proxy_renewal td: %d, threshold: %d" % (td_in_seconds, config.vm_proxy_renewal_threshold))
        return td_in_seconds < config.vm_proxy_renewal_threshold

    def needs_proxy_shutdown(self):
        """This method will test if a VM needs to be shutdown before proxy expiry, according
        the VM proxy shutdown threshold found in the cloud scheduler configuration.

        Returns: True if the VM needs to be shutdown
                 False otherwise (or if the VM has no user proxy associated with it).
        """
        expiry_time = self.get_x509userproxy_expiry_time()
        if expiry_time == None:
            return False
        td = expiry_time - datetime.datetime.utcnow()
        td_in_seconds = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
        log.verbose("needs_proxy_renewal td: %d, threshold: %d" % (td_in_seconds, config.vm_proxy_shutdown_threshold))
        return td_in_seconds < config.vm_proxy_shutdown_threshold

    def get_env(self):
        """The following method will return the environment that should
        be used when executing subprocesses.  This is needed for setting
        the user's x509 proxy for example.
        """
        env = None
        if self.get_proxy_file() != None:
            env = {'X509_USER_PROXY':self.get_proxy_file()}
        return env

class NoResourcesError(Exception):
    """Exception raised for errors where not enough resources are available

    Attributes:
        resource -- name of resource that is insufficient

    """

    def __init__(self, resource):
        self.resource = resource


class ICluster:
    """
    The ICluster interface is the framework for implementing support for
    a specific IaaS cloud implementation. In general, you'll need to
    override __init__ (be sure to call super's init), vm_create, vm_poll,
    and vm_destroy
    """

    def __init__(self, name="Dummy Cluster", host="localhost",
                 cloud_type="Dummy", memory=[], max_vm_mem= -1, networks=[],
                 vm_slots=0, cpu_cores=0, storage=0, hypervisor='xen', boot_timeout=None, enabled=True, priority=0):
        self.name = name
        self.network_address = host
        self.cloud_type = cloud_type
        self.memory = memory
        self.max_mem = tuple(memory)
        self.max_vm_mem = max_vm_mem
        self.network_pools = networks
        self.vm_slots = vm_slots
        self.max_slots = vm_slots
        self.cpu_cores = cpu_cores
        self.storageGB = storage
        self.max_storageGB = storage
        self.vms = [] # List of running VMs
        self.vms_lock = threading.RLock()
        self.res_lock = threading.RLock()
        self.enabled = enabled
        self.hypervisor = hypervisor
        self.boot_timeout = int(boot_timeout) if boot_timeout != None else config.vm_start_running_timeout
        self.connection_fail_disable_time = config.connection_fail_disable_time
        self.connection_problem = False
        self.errorconnect = None
        self.priority = priority
        self.failed_image_set = set()

        self.setup_logging()
        log.debug("New cluster %s created" % self.name)

    def __getstate__(self):
        """Override to work with pickle module."""
        state = self.__dict__.copy()
        del state['vms_lock']
        del state['res_lock']
        return state

    def __setstate__(self, state):
        """Override to work with pickle module."""
        self.__dict__ = state
        self.vms_lock = threading.RLock()
        self.res_lock = threading.RLock()

    def __repr__(self):
        return self.name

    def setup_logging(self):
        """Fetch the global log object."""
        global log
        log = logging.getLogger("cloudscheduler")


    def log_cluster(self):
        """Print cluster information to the log."""
        log.info("-" * 30 +
            "Name:\t\t%s\n"        % self.name +
            "Address:\t%s\n"       % self.network_address +
            "Type:\t\t%s\n"        % self.cloud_type +
            "VM Slots:\t%s\n"      % self.vm_slots +
            "CPU Cores:\t%s\n"     % self.cpu_cores +
            "Storage:\t%s\n"       % self.storageGB +
            "Memory:\t\t%s\n"      % self.memory +
            "Network Pools:\t%s\n" % string.join(self.network_pools, ", ") +
            "-" * 30)

    def log(self):
        """Print a short form of cluster information to the log."""
        log.debug("CLUSTER Name: %s, Address: %s, Type: %s, VM slots: %d, Mem: %s" \
          % (self.name, self.network_address, self.cloud_type, self.vm_slots, \
          self.memory))

    def log_vms(self):
        """Print the cluster 'vms' list (via VM print)."""
        if len(self.vms) == 0:
            log.info("CLUSTER %s has no running VMs..." % (self.name))
        else:
            log.info("CLUSTER %s running VMs:" % (self.name))
            for vm in self.vms:
                vm.log_short("\t")


    ## Support methods

    def num_vms(self):
        """Returns the number of VMs running on the cluster (in accordance
        to the vms[] list)
        """
        return len(self.vms)

    def slot_fill_ratio(self):
        """Return a ratio of how 'full' the cluster is based on used slots / total slots."""
        return (self.max_slots - self.vm_slots) / float(self.max_slots)

    def get_cluster_info_short(self):
        """Return a short form of cluster information."""
        output = "Cluster: {0} \n".format(self.name)
        output += "{0:25}  {1:^15}  {2:^15}  {3:^12} {4:^15} {5:^12} {6:^12}\n".format("ADDRESS", "CLOUD TYPE", "VM SLOTS", "MEMORY", "STORAGE", "PRIORITY", "ENABLED")
        output += "{0:^25}  {1:^15}  {2:^15}  {3:^12} {4:^15} {5:^12} {6:^12}\n".format(self.network_address[0:24], self.cloud_type, self.vm_slots, self.memory, self.storageGB, self.priority, str(self.enabled))
        return output

    def get_cluster_vms_info(self):
        """Return information about running VMs on Cluster as a string."""
        if len(self.vms) == 0:
            return ""
        else:
            output = ""
            for vm in self.vms:
                output += "%s %-15s\n" % (vm.get_vm_info()[:-1], self.name)
            return output

    def get_vm(self, vm_id):
        """Get VM object with id value."""
        for vm in self.vms:
            if vm_id == vm.id:
                return vm
        return None


    # VM manipulation methods
    #-!------------------------------------------------------------------------
    # NOTE: In implementing subclasses of Cluster, the following method prototypes
    #       should be used (standardize on these parameters)
    #-!------------------------------------------------------------------------

    # Note: vm_id is the identifier for a VM, used to query or change an already
    #       created VM. vm_id will be a different entity based on the subclass's
    #       cloud software. EG:
    #       - Nimbus vm_ids are epr files
    #       - OpenNebula (and Eucalyptus?) vm_ids are names/numbers

    def vm_create(self, **args):
        log.debug('This method should be defined by all subclasses of Cluster\n')
        assert 0, 'Must define workspace_create'

    def vm_destroy(self, vm, return_resources=True, reason=""):
        log.debug('This method should be defined by all subclasses of Cluster\n')
        assert 0, 'Must define workspace_destroy'

    def vm_poll(self, vm):
        log.debug('This method should be defined by all subclasses of Cluster\n')
        assert 0, 'Must define workspace_poll'


    ## Private VM methods

    def find_mementry(self, memory):
        """Finds a memory entry in the Cluster's 'memory' list which supports the
        requested amount of memory for the VM. If multiple memory entries fit
        the request, returns the first suitable entry. Returns an exact fit if
        one exists.
        Parameters: memory - the memory required for VM creation
        Return: The index of the first fitting entry in the Cluster's 'memory'
        list.
        If no fitting memory entries are found, returns -1 (error!)
        """
        # Check for exact fit
        if (memory in self.memory):
            return self.memory.index(memory)

        # Scan for any fit
        for i in range(len(self.memory)):
            if self.memory[i] >= memory:
                return i

        # If no entries found, return error code.
        return(-1)

    def find_potential_mementry(self, memory):
        """Check if a cluster contains a memory entry with adequate space for given memory value.
        Returns: True if a valid memory entry is found
                 False otherwise
        """
        potential_fit = False
        for i in range(len(self.max_mem)):
            if self.max_mem[i] >= memory:
                potential_fit = True
                break
        return potential_fit

    def resource_checkout(self, vm):
        """
        Checks out resources taken by a VM in creation from the internal rep-
        resentation of the Cluster

        Parameters:
        vm   - the VM object used to check out resources from the Cluster.

        Raises NoResourcesError if there are not enough available resources
        to check out.
        """
        #log.debug("Checking out resources for VM %s from Cluster %s" % (vm.name, self.name))
        with self.res_lock:

            remaining_vm_slots = self.vm_slots - 1
            if remaining_vm_slots < 0:
                raise NoResourcesError("vm_slots")

            remaining_storage = self.storageGB - vm.storage
            if remaining_storage < 0:
                raise NoResourcesError("storage")

            remaining_memory = self.memory[vm.mementry] - vm.memory
            if remaining_memory < 0:
                raise NoResourcesError("memory")

            # Otherwise, we can check out these resources
            self.vm_slots = remaining_vm_slots
            self.storageGB = remaining_storage
            self.memory[vm.mementry] = remaining_memory

    def resource_return(self, vm):
        """Returns the resources taken by the passed in VM to the Cluster's internal
        storage.
        Parameters: (as for checkout() )
        Notes: (as for checkout)
        """
        #log.debug("Returning resources used by VM %s to Cluster %s" % (vm.name, self.name))
        with self.res_lock:
            self.vm_slots += 1
            self.storageGB += vm.storage
            # ISSUE: No way to know what mementry a VM is running on
            try:
                self.memory[vm.mementry] += vm.memory
            except:
                log.warning("Couldn't return memory because I don't know about that mem entry anymore...")




import os
import sys
import time
import uuid
import string
import shutil
import logging
import nimbus_xml
import subprocess
import cluster_tools
import cloud_init_util
import cloudscheduler.config as config
import cloudscheduler.utilities as utilities
from cloudscheduler.job_management import _attr_list_to_dict

log = utilities.get_cloudscheduler_logger()

class OpenStackCluster(cluster_tools.ICluster):
    ERROR = 1
    DEFAULT_INSTANCE_TYPE = config.default_VMInstanceType if config.default_VMInstanceType else "m1.small"
    DEFAULT_INSTANCE_TYPE_LIST = _attr_list_to_dict(config.default_VMInstanceTypeList)
    VM_STATES = {
            "BUILD" : "Starting",
            "ACTIVE" : "Running",
            "SHUTOFF" : "Shutdown",
            "SUSPENDED": "Suspended",
            "PAUSED": "Paused",
            "ERROR" : "Error",
    }
    def __init__(self, name="Dummy Cluster", cloud_type="Dummy",
                 memory=[], max_vm_mem= -1, networks=[], vm_slots=0,
                 cpu_cores=0, storage=0, security_group=None,
                 username=None, password=None, tenant_name=None, auth_url=None,
                 hypervisor='xen', key_name=None, boot_timeout=None, secure_connection="",
                 regions=[], vm_domain_name="", reverse_dns_lookup=False,placement_zone=None, 
                 enabled=True, priority=0, cacert=None):

        # Call super class's init
        cluster_tools.ICluster.__init__(self,name=name, host=auth_url, cloud_type=cloud_type,
                         memory=memory, max_vm_mem=max_vm_mem, networks=networks,
                         vm_slots=vm_slots, cpu_cores=cpu_cores,
                         storage=storage, hypervisor=hypervisor, boot_timeout=boot_timeout, enabled=enabled, priority=priority)
        try:
            import novaclient.v1_1.client as nvclient
            import novaclient.exceptions
            #import keystoneclient.v2_0.client as ksclient
        except:
                print "Unable to import novaclient - cannot use native openstack cloudtypes"
                sys.exit(1)
        if not security_group:
            security_group = ["default"]
        self.security_groups = security_group
        self.username = username if username else ""
        self.password = password if password else ""
        self.tenant_name = tenant_name if tenant_name else ""
        self.auth_url = auth_url if auth_url else ""
        self.key_name = key_name if key_name else ""
        self.secure_connection = secure_connection in ['True', 'true', 'TRUE']
        self.total_cpu_cores = -1
        self.regions = regions
        self.vm_domain_name = vm_domain_name if vm_domain_name != None else ""
        self.reverse_dns_lookup = reverse_dns_lookup in ['True', 'true', 'TRUE']
        self.placement_zone = placement_zone
        self.flavor_set = set()
        self.cacert = cacert
    
    def __getstate__(self):
        """Override to work with pickle module."""
        state = cluster_tools.ICluster.__getstate__(self)
        del state['flavor_set']
        return state

    def __setstate__(self, state):
        """Override to work with pickle module."""
        cluster_tools.ICluster.__setstate__(self, state)
        self.flavor_set = set()
    
    def vm_create(self, vm_name, vm_type, vm_user, vm_networkassoc,
                  vm_image, vm_mem, vm_cores, vm_storage, customization=None,
                  vm_keepalive=0, instance_type="", job_per_core=False, 
                  securitygroup=[],key_name="", pre_customization=None, use_cloud_init=False, extra_userdata=[]):
        """ Create a VM on OpenStack."""

        import novaclient.exceptions
        use_cloud_init = use_cloud_init or config.use_cloud_init
        nova = self._get_creds_nova()
        if len(securitygroup) != 0:
            sec_group = []
            for group in securitygroup:
                if group in self.security_groups:
                    sec_group.append(group)
            if len(sec_group) == 0:
                log.debug("No defined security groups for job - trying default value from cloud_resources.conf")
                sec_group = self.security_groups
        else:
            sec_group = self.security_groups
        log.debug("Using security group: %s" % str(sec_group))
        if key_name and len(key_name) > 0:
            if not nova.keypairs.findall(name=key_name):
                key_name = ""
        else:
            key_name = self.key_name if self.key_name else ""
        if customization:
            if not use_cloud_init:
                user_data = nimbus_xml.ws_optional(customization)
            else:
                user_data = cloud_init_util.build_write_files_cloud_init(customization)
        else:
            user_data = ""
        if pre_customization:
            if not use_cloud_init:
                for item in pre_customization:
                    user_data = '\n'.join([item, user_data])
            else:
                user_data = cloud_init_util.inject_customizations(pre_customization, user_data)
        elif use_cloud_init:
            user_data = cloud_init_util.inject_customizations([], user_data)
        if len(extra_userdata) > 0:
            # need to use the multi-mime type functions
            user_data = cloud_init_util.build_multi_mime_message([(user_data, 'cloud-config', 'cloud_conf.yaml')], extra_userdata)
        
        try:
            image = vm_image[self.name]
        except:
            try:
                image = vm_image[self.network_address]
            except:
                try:
                    vm_default_ami = _attr_list_to_dict(config.default_VMAMI)
                    if self.name in vm_default_ami.keys():
                        image = vm_default_ami[self.name]
                    else:
                        image = vm_default_ami[self.network_address]
                except:
                    try:
                        image = vm_default_ami["default"]
                    except:
                        log.exception("Can't find a suitable AMI")
                        return
        try:
            imageobj = nova.images.find(name=image)
        except Exception as e:
            log.warning("Exception occurred while trying to fetch image via name: %s" % e)
            try:
                imageobj = nova.images.get(image)
                log.debug("Got image via uuid: %s" % image)
            except Exception as e:
                log.exception("Unable to fetch image via uuid: %s" % e)
                self.failed_image_set.add(image)
                return
        try:
            if self.name in instance_type.keys():
                i_type = instance_type[self.name]
            else:
                i_type = instance_type[self.network_address]
        except:
            log.debug("No instance type for %s, trying default" % self.network_address)
            try:
                if self.name in self.DEFAULT_INSTANCE_TYPE_LIST.keys():
                    i_type = self.DEFAULT_INSTANCE_TYPE_LIST[self.name]
                else:
                    i_type = self.DEFAULT_INSTANCE_TYPE_LIST[self.network_address]
            except:
                log.debug("No default instance type found for %s, trying single default" % self.network_address)
                i_type = self.DEFAULT_INSTANCE_TYPE
        try:   
            flavor = nova.flavors.find(name=i_type)
        except Exception as e:
            log.warning("Exception occurred while trying to get flavor by name: %s - will attempt to use name value as a uuid." % e)
            try:
                flavor = nova.flavors.get(i_type)
                log.debug("Got flavor via uuid: %s" % i_type)
            except Exception as ex:
                log.error("Exception occurred trying to get flavor by uuid: %s" % ex)
                return
        self.flavor_set.add(flavor)
        # find the network id to use if more than one network
        if vm_networkassoc:
            network = self._find_network(vm_networkassoc)
            if network:
                netid = [{'net-id': network.id}]
            else:
                log.debug("Unable to find network named: %s on %s" % (vm_networkassoc, self.name))
                netid = []
        elif self.network_pools and len(self.network_pools) > 0:
            network = self._find_network(self.network_pools[0])
            if network:
                netid = [{'net-id': network.id}]
            else:
                log.debug("Unable to find network named: %s on %s" % (self.network_pools[0], self.name))
                netid = []
        else:
            netid = []
        # Need to get the rotating hostname from the google code to use for here.  
        name = self._generate_next_name()
        instance = None

        if name:
            try:
                instance = nova.servers.create(name=name, image=imageobj, flavor=flavor, key_name=key_name, 
                                               availability_zone=self.placement_zone, nics =netid, userdata=user_data, security_groups=sec_group)
                #print instance.__dict__
            except novaclient.exceptions.OverLimit as e:
                log.info("Unable to create VM without exceeded quota on %s: %s" % (self.name, e.message))
            except Exception as e:
                #print e
                log.error("Unhandled exception while creating vm on %s: %s" %(self.name, e))
            if instance:
                instance_id = instance.id
                
                new_vm = cluster_tools.VM(name = vm_name, id = instance_id, vmtype = vm_type, user = vm_user,
                            clusteraddr = self.network_address, hostname = ''.join([name, self.vm_domain_name]),
                            cloudtype = self.cloud_type, network = vm_networkassoc,
                            image= vm_image, flavor=flavor.name,
                            memory = vm_mem, cpucores = vm_cores, storage = vm_storage, 
                            keep_alive = vm_keepalive, job_per_core = job_per_core)
    
                try:
                    self.resource_checkout(new_vm)
                except:
                    log.error("Unexpected Error checking out resources when creating a VM. Programming error?")
                    self.vm_destroy(new_vm, reason="Failed Resource checkout")
                    return self.ERROR
        
                self.vms.append(new_vm)
            else:
                log.debug("Failed to create instance on %s" % self.name)
                return self.ERROR
        else:
            log.debug("Unable to generate name for %" % self.name)
            return self.ERROR

        return 0

    def vm_destroy(self, vm, return_resources=True, reason=""):
        """ Destroy a VM on OpenStack."""
        nova = self._get_creds_nova()
        import novaclient.exceptions
        log.info("Destroying VM: %s Name: %s Reason: %s" % (vm.id, vm.hostname, reason))
        try:
            instance = nova.servers.get(vm.id)
            instance.delete()
        except novaclient.exceptions.NotFound as e:
            log.error("VM %s not found on %s: removing from CS" % (vm.id, self.name))
        except Exception as e:
            try:
                log.error("Unhandled exception while destroying VM on %s : %s" % (self.name,e))
                return 1
            except:
                log.error("Failed to log exception properly?")
                return 1
            

        # Delete references to this VM
        try:
            if return_resources and vm.return_resources:
                self.resource_return(vm)
            with self.vms_lock:
                self.vms.remove(vm)
        except Exception as e:
            log.error("Error removing vm from list: %s" % e)
            return 1

        return 0
    def vm_poll(self, vm):
        """ Query OpenStack for status information of VMs."""
        import novaclient.exceptions
        nova = self._get_creds_nova()
        instance = None
        try:
            instance = nova.servers.get(vm.id)
        except novaclient.exceptions.NotFound as e:
            log.exception("VM %s not found on %s: %s" % (vm.id, self.name, e))
            vm.status = self.VM_STATES['ERROR']
        except Exception as e:
            try:
                log.error("Unexpected exception occurred polling vm %s: %s" % (vm.id, e))
            except:
                log.error("Failed to log exception properly: %s" % vm.id)
        with self.vms_lock:
            #print instance.status
            if instance and vm.status != self.VM_STATES.get(instance.status, "Starting"):

                vm.last_state_change = int(time.time())
                log.debug("VM: %s on %s. Changed from %s to %s." % (vm.id, self.name, vm.status, self.VM_STATES.get(instance.status, "Starting")))
            if instance and instance.status in self.VM_STATES.keys():
                vm.status = self.VM_STATES[instance.status]
            elif instance:
                vm.status = instance.status
            else:
                vm.status = self.VM_STATES['ERROR']
        return vm.status

    def _get_creds_nova(self):
        """Get an auth token to Nova."""
        try:
            import novaclient.v1_1.client as nvclient
        except:
                print "Unable to import novaclient - cannot use native openstack cloudtypes"
                sys.exit(1)
        try:
            client = nvclient.Client(username=self.username, api_key=self.password, auth_url=self.auth_url, project_id=self.tenant_name,
                                    region_name=self.regions[0], cacert=self.cacert)
        except Exception as e:
            log.error("Unable to create connection to %s: Reason: %s" % (self.name, e))
        return client 

    def _generate_next_name(self):
        name = ''.join([self.name.replace('_', '-').lower(), '-', str(uuid.uuid4())])
        collision = False
        for vm in self.vms:
            if name == vm.hostname:
                collision= True
                break
        if collision:
            name = None
        return name
    
    def _find_network(self, name):
        nova = self._get_creds_nova()
        network = None
        try:
            networks = nova.networks.list()
            for net in networks:
                if net.label == name:
                    network = net
        except Exception as e:
            log.error("Unable to list networks on %s Exception: %s" % (self.name, e))
        return network




# setup.py - standard distutils setup file for Cloud Scheduler
import sys
try:
    from setuptools import setup
except:
    try:
        from distutils.core import setup
    except:
        print "Couldn't use either setuputils or distutils. Install one of those. :)"
        sys.exit(1)
import cloudscheduler.__version__ as version

setup(name = "cloud-scheduler",
    version = version.version,
    license="'GPL3' or 'Apache 2'",
    install_requires=[
       "boto>=2.0b3",
        ],
    description = "A cloud-enabled distributed resource manager",
    author = "Duncan Penfold-Brown, Chris Usher, Patrick Armstrong, Ian Gable, Michael Paterson, Andre Charbonneau",
    author_email = "mhp@uvic.ca",
    url = "http://github.com/hep-gc/cloud-scheduler",
    packages = ['cloudscheduler'],
    package_data = {'cloudscheduler' : ["wsdl/*"] },
    data_files = [("share/cloud-scheduler/", ["cloud_scheduler.conf", "cloud_resources.conf", "scripts/cloud_scheduler.init.d", "scripts/cloud_scheduler.sysconf"])],
    scripts = ["cloud_scheduler", "cloud_status", "cloud_admin"],
) 



import os
import threading
import time
import traceback
import datetime
import subprocess
import tempfile
import shutil
import shlex

import cloudscheduler.config as config
import cloudscheduler.utilities as utilities
import cloudscheduler.job_management as job_management
import cloudscheduler.cloud_management as cloud_management

from cloudscheduler.cluster_tools import VM
from job_management import Job

log = utilities.get_cloudscheduler_logger()

class JobProxyRefresher(threading.Thread):
    """
    JobProxyRefresher - Periodically checks the expiry time on job user proxies and attempt
                        to renew the ones about to expire using MyProxy.
    """


    def __init__(self, job_pool):
        threading.Thread.__init__(self, name=self.__class__.__name__)
        self.job_pool = job_pool
        self.quit = False
        self.heart_beat = time.time()
        self.polling_interval = config.job_proxy_refresher_interval # proxy expiry time poll interval, in seconds

    def stop(self):
        log.debug("Waiting for job proxy refresher loop to end")
        self.quit = True

    def run(self):
        try:
            log.info("Starting JobProxyRefresher thread...")

            while not self.quit:
                # The following timestamp is use to time this proxy refresh cycle.
                cycle_start_ts = datetime.datetime.today()

                jobs = self.job_pool.job_container.get_all_jobs()
                log.verbose("Refreshing job user proxies. [%d proxies to process]" % (len(jobs)))
                for job in jobs:
                    jobcertextime = job.get_x509userproxy_expiry_time()
                    if jobcertextime:
                        log.verbose("Proxy for job %s expires in %s" % (job.id, jobcertextime - datetime.datetime.utcnow()))
                    if job.is_proxy_expired():
                        log.warning("Proxy for job %s is expired.  Skipping proxy renewal for this job." % (job.id))
                    elif job.needs_proxy_renewal():
                        if job.get_myproxy_creds_name() != None:
                            log.verbose("Renewing proxy %s for job %s" % (job.get_x509userproxy(), job.id))
                            if MyProxyProxyRefresher().renew_proxy(job.get_x509userproxy(), job.get_myproxy_creds_name(), job.get_myproxy_server(), job.get_myproxy_server_port(), job.get_renew_time()):
                                # Yay, proxy renewal worked! :-)
                                log.verbose("Proxy for job %s renewed." % (job.id))
                                # Don't forget to reset the proxy expiry time cache.
                                job.reset_x509userproxy_expiry_time()
                            else:
                                log.error("Error renewing proxy for job %s" % (job.id))
                        else:
                            # If we get here, this means that the proxy should be renewed, but there
                            # is not MyProxy info for that job's proxy.  Not an error; just that the
                            # owner of the job didn't give any MyProxy information to renew the
                            # credentials.
                            log.debug("Not renewing proxy for job %s because missing MyProxy info." % (job.id))
                    else:
                        log.verbose("No need to renew proxy for job %s" % (job.id))

                # Lets record the current time and then log how much time the cycle took.
                cycle_end_ts = datetime.datetime.today()
                log.verbose("Job proxy refreshing cycle done. [%d;%s;%s;%s]" % (len(jobs), cycle_start_ts, cycle_end_ts, cycle_end_ts - cycle_start_ts))

                log.verbose("JobProxyRefresher waiting %ds..." % self.polling_interval)
                sleep_tics = self.polling_interval
                self.heart_beat = time.time()
                while (not self.quit) and sleep_tics > 0:
                    time.sleep(1)
                    sleep_tics -= 1

            log.info("Exiting JobProxyRefresher thread")
        except:
            log.error("Error in JobProxyRefresher thread.")
            log.error(traceback.format_exc())

class VMProxyRefresher(threading.Thread):
    """
    VMProxyRefresher - Periodically checks the expiry time on VM user proxies and attempt
                        to renew the ones about to expire using MyProxy.
    """


    def __init__(self, cloud_resources):
        threading.Thread.__init__(self, name=self.__class__.__name__)
        self.cloud_resources = cloud_resources
        self.quit = False
        self.heart_beat = time.time()
        self.polling_interval = config.vm_proxy_refresher_interval # proxy expiry time poll interval, in seconds

    def stop(self):
        log.debug("Waiting for VM proxy refresher loop to end")
        self.quit = True

    def run(self):
        try:
            log.info("Starting VMProxyRefresher thread...")

            while not self.quit:
                # The following timestamp is use to time this proxy refresh cycle.
                cycle_start_ts = datetime.datetime.today()

                vms = self.cloud_resources.get_all_vms()
                log.verbose("Refreshing VM proxies. [%d proxies to process]" % (len(vms)))
                for vm in vms:
                    vmcertextime = vm.get_x509userproxy_expiry_time()
                    if vmcertextime:
                        log.verbose("Proxy for VM %s expires in %s" % (vm.id, vmcertextime - datetime.datetime.utcnow()))
                    if vm.is_proxy_expired():
                        log.warning("Proxy for VM %s is expired.  Skipping proxy renewal for this VM." % (vm.id))
                    elif vm.needs_proxy_renewal():
                        if vm.get_myproxy_creds_name() != None:
                            log.debug("Renewing proxy %s for VM %s" % (vm.get_proxy_file(), vm.id))
                            if MyProxyProxyRefresher().renew_proxy(vm.get_proxy_file(), vm.get_myproxy_creds_name(), vm.get_myproxy_server(), vm.get_myproxy_server_port(), vm.get_renew_time()):
                                # Yay, proxy renewal worked! :-)
                                log.verbose("Proxy for VM %s renewed." % (vm.id))
                                # Don't forget to reset the proxy expiry time cache.
                                vm.reset_x509userproxy_expiry_time()
                            else:
                                log.error("Error renewing proxy for VM %s" % (vm.id))
                        else:
                            # If we get here, this means that the proxy should be renewed, but there
                            # is no MyProxy info for that vm's proxy.  Not an error; just that the
                            # owner of the vm didn't give any MyProxy information to renew the
                            # credentials.
                            log.debug("Not renewing proxy for VM %s because missing MyProxy info." % (vm.id))
                    else:
                        log.verbose("No need to renew proxy for VM %s" % (vm.id))

                # Lets record the current time and then log how much time the cycle took.
                cycle_end_ts = datetime.datetime.today()
                log.verbose("VM proxy refreshing cycle done. [%d;%s;%s;%s]" % (len(vms), cycle_start_ts, cycle_end_ts, cycle_end_ts - cycle_start_ts))

                log.verbose("VMProxyRefresher waiting %ds..." % self.polling_interval)
                sleep_tics = self.polling_interval
                self.heart_beat = time.time()
                while (not self.quit) and sleep_tics > 0:
                    time.sleep(1)
                    sleep_tics -= 1

            log.info("Exiting VMProxyRefresher thread")
        except:
            log.error("Error in VMProxyRefresher thread.")
            log.error(traceback.format_exc())



class MyProxyProxyRefresher():
    """
    Utility class used to refresh a proxy using a MyProxy server.
    """
    def renew_proxy(self, proxy_file_path, myproxy_creds_name, myproxy_server, myproxy_server_port, renew_time):
        """This method will call the MyProxy commands to renew the credential for a given job.
        
        Returns True on sucess, False otherwise."""
        if proxy_file_path == None:
            log.error("Attempt to renew proxy for job with no proxy.  Aborting proxy renew operation.")
            return False

        if myproxy_creds_name == None:
            log.error("Missing MyProxy credential name for job.")
            return False

        if myproxy_server == None:
            log.warning("MyProxy credential name given but missing MyProxy server host. Defaulting to localhost")
            myproxy_server = "localhost"

        if myproxy_server_port == None:
            myproxy_server_port = "7512"

        myproxy_command = config.myproxy_logon_command
        if not os.path.isabs(myproxy_command):
            try:
                myproxy_command = utilities.get_globus_path(executable=myproxy_command) + myproxy_command
            except:
                log.exception("Problem getting myproxy-logon path")
                return False


        # Note: Here we put the refreshed proxy in a seperate file.  We do this to protect the ownership and permisions on the
        # original user's proxy in case the condor_submit was run on the same machine as the cloud scheduler.  If we do not do
        # this, then the cloud scheduler will overwrite the user's proxy with the renewed proxy, but will be owned by root; so
        # user will not be able to create proxies anymore because he/she won't have permission to overwrite the proxy file
        # created by the cloud scheduler.
        # Once the renewed proxy is in the temporary file, we then copy it over the original job's proxy, without changing its
        # permissions.
        #
        # This is a bit of a hack; there must me a better way to handle this problem.
        try:
            (new_proxy_file, new_proxy_file_path) = tempfile.mkstemp(suffix='.csRenewedProxy')
            os.close(new_proxy_file)
            myproxy_logon_cmd = '%s -s %s -p %s -k "%s" -a %s -o %s -t %s -d -v' % (myproxy_command, myproxy_server, myproxy_server_port, myproxy_creds_name, proxy_file_path, new_proxy_file_path, renew_time)
            cmd_args = shlex.split(myproxy_logon_cmd)
            log.verbose('Invoking myproxy-logon command to refresh proxy %s ...' % (proxy_file_path))
            myproxy_logon_process = subprocess.Popen(cmd_args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (stdout, stderr) = myproxy_logon_process.communicate()
            if myproxy_logon_process.returncode != 0:
                log.error("Error renewing proxy from MyProxy server: %s %s" % (stdout, stderr))
                os.remove(new_proxy_file_path)
                return False
            else:
                log.verbose('myproxy-logon command returned successfully')

            shutil.copyfile(new_proxy_file_path, proxy_file_path)
            os.remove(new_proxy_file_path)
        except:
            log.exception('Unexpected error during proxy renewal using MyProxy.')
            return False

        return True
    
    def renew_proxy_meta(self, joborvm):
        """Single function to accept a job or vm proxy to refresh."""
        if joborvm is isinstance(Job):
            if MyProxyProxyRefresher().renew_proxy(joborvm.get_x509userproxy(), joborvm.get_myproxy_creds_name(), joborvm.get_myproxy_server(), joborvm.get_myproxy_server_port()):
                # Yay, proxy renewal worked! :-)
                log.verbose("Proxy for job %s renewed." % (joborvm.id))
                # Don't forget to reset the proxy expiry time cache.
                joborvm.reset_x509userproxy_expiry_time()
            else:
                log.error("Error renewing proxy for job %s" % (joborvm.id))
        elif joborvm is isinstance(VM):
            if self.renew_proxy(joborvm.get_proxy_file(), joborvm.get_myproxy_creds_name(), joborvm.get_myproxy_server(), joborvm.get_myproxy_server_port()):
                # Yay, proxy renewal worked! :-)
                log.verbose("Proxy for VM %s renewed." % (joborvm.id))
                # Don't forget to reset the proxy expiry time cache.
                joborvm.reset_x509userproxy_expiry_time()
            else:
                log.error("Error renewing proxy for VM %s" % (joborvm.id))

    def renew_job_proxy_user(self, job_pool, user):
        """Refresh all job proxies for a user."""
        for job in job_pool.job_container.get_jobs_for_user(user):
            self.renew_proxy_meta(job)

    def renew_vm_proxy_user(self, resource_pool, user):
        """Refresh all VM proxies for a user."""
        for vm in resource_pool.get_user_vms():
            self.renew_proxy_meta(vm)




import cluster_tools
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.Runner import Runner
import time
import os
import ConfigParser
import threading
from stratuslab.Exceptions import OneException
import cloudscheduler.utilities as utilities
import base64

log = utilities.get_cloudscheduler_logger()


class StratusLabCluster(cluster_tools.ICluster):
    
    VM_TARGETSTATE = "Running"
    VM_NODES = "1"
    VM_SHUTDOWN = 90.0
    ERROR = 1
    
    VM_STATES = {
                 'INIT'      : 'Starting',
                 'BOOT'      : 'Starting',
                 'PROLOG'    : 'Starting',
                 'PENDING'   : 'Starting',
                 'HOLD'      : 'Starting',
                 'RUNNING'   : 'Running',
                 'ACTIVE'    : 'Running',
                 'STOPPED'   : 'Running',
                 'SUSPENDED' : 'Running',
                 'DONE'      : 'Shutdown',
                 'EPILOG'    : 'Shutdown',
                 'FAILED'    : 'Error',
                 'FAILURE'   : 'Error',
                 'UNKNOWN'   : 'Error',
    }

    config = ConfigParser.ConfigParser()    
    config_file = os.path.expanduser('~/.stratuslab/stratuslab-user.cfg')
    config_section = 'default'
    config.read(config_file)

    #Read the stratuslab configuration file
    username, password, endpoint = None, None, None         
    
    try: 
        username = config.get(config_section, 'username')
    except ConfigParser.NoOptionError: 
        log.error("Stratuslab config file %s does not contain username. VM creation will fail" %(config_file))
    except ConfigParser.NoSectionError: 
        log.error("No section '%s' found Stratuslab config file '%s'. Do the file and section exist? VM creation will fail" %(config_section, config_file))

    try: 
        password = config.get(config_section, 'password')
    except ConfigParser.NoOptionError: 
        log.error("Stratuslab config file %s does not contain password. VM creation will fail" %(config_file))
    except ConfigParser.NoSectionError: 
        log.error("No section '%s' found Stratuslab config file '%s'. Do the file and section exist? VM creation will fail" %(config_section, config_file))

    try: 
        endpoint = config.get(config_section, 'endpoint')
    except ConfigParser.NoOptionError: 
        log.error("Stratuslab config file %s does not contain endpoint. VM creation will fail" %(config_file))
    except ConfigParser.NoSectionError: 
        log.error("No section '%s' found Stratuslab config file '%s'. Do the file and section exist? VM creation will fail" %(config_section, config_file))
    
    options = Runner.defaultRunOptions()
    options.update({'username'            : username,
                    'password'            : password,
                    'endpoint'            : endpoint,
                    'marketplaceEndpoint' : 'https://marketplace.stratuslab.eu',                    
                    'mpi_machine_file'    : True, 
                    #'instanceType'        : self.instanceType,
                    'cluster_admin'       : 'root', 
                    'master_vmid'         : None,
                    'tolerate_failures'   : False, 
                    'clean_after_failure' : False,
                    'include_master'      : True, 
                    'shared_folder'       :'/home',
                    'add_packages'        : None, 
                    'ssh_hostbased'       : True, 
                    #'instanceNumber'      : 0,
                    'verboseLevel'        : 0
                    })
    _v_configHolder = ConfigHolder(options)

    __idMap = {}
    __vmMap = {}

    
    def __init__(self, name = "Dummy StratusLab Cluster", host = "localhost", cloud_type = "StratusLab",
                 memory = [], max_vm_mem = -1, cpu_archs = [], networks = [], vm_slots = 0,
                 cpu_cores = 0, storage = 0, access_key_id = None, secret_access_key = None,
                 hypervisor = 'xen', key_name = None, vm_loc = '', contextualization = '', enabled=True, priority=0):

        # Call super class' init
        cluster_tools.ICluster.__init__(self, name = name, host = host, cloud_type = cloud_type,
                         memory = memory, max_vm_mem = max_vm_mem, cpu_archs = cpu_archs, networks = networks,
                         vm_slots = vm_slots, cpu_cores = cpu_cores, storage = storage, hypervisor = hypervisor,
                         enabled=enabled, priority=priority)

        try:
            f = open(contextualization, 'r')
            s = f.read()
            StratusLabCluster._v_configHolder.set('extraContextData', 'EC2_USER_DATA=%s' %base64.standard_b64encode(s))
        except IOError:
            log.error("Contextualization file '%s' is not valid. Proceeding without contextualization..." % str(contextualization) )

        self.__runnerIds = {}
     
            
    def vm_create(self, vm_name, vm_type = "CernVM", vm_user = "root", vm_networkassoc = "", vm_cpuarch = "",
            vm_image = "", vm_mem = 1, vm_cores = 1, vm_storage = 30, customization = None, vm_keepalive = 0,
            job_proxy_file_path = None, myproxy_creds_name = None, myproxy_server = None, 
            myproxy_server_port = None, job_per_core = False, proxy_non_boot = False,
            vmimage_proxy_file = None, vmimage_proxy_file_path = None, vm_loc = ''):

        log.debug("Running new instance with Marketplace id %s in StratusLab" % str(vm_loc))
        runner = None
        
        if vm_loc not in StratusLabCluster.__idMap:
            runner = Runner(vm_loc, StratusLabCluster._v_configHolder) #vm_loc: URL of VM or key? Does not seem to matter in Runner (l. 506)
            StratusLabCluster.__idMap[vm_loc] = runner
        else:
            runner = StratusLabCluster.__idMap[vm_loc]
        
        try:
            ids = runner.runInstance()
            log.debug("Created instances: %s" % str(ids))
            #for new_id in ids:
            new_id = ids[len(ids) - 1]
            new_vm = cluster_tools.VM(name = vm_name, id = str(new_id), vmtype = vm_type, user = vm_user,
                network = vm_networkassoc, cpuarch = vm_cpuarch, image = vm_image, memory = vm_mem, 
                cpucores = vm_cores, storage = vm_storage, keep_alive = vm_keepalive, 
                myproxy_creds_name = myproxy_creds_name, myproxy_server = myproxy_server, 
                myproxy_server_port = myproxy_server_port, job_per_core = job_per_core)
            
            StratusLabCluster.__vmMap[str(new_id)] = vm_loc
            
            if vm_loc not in self.__runnerIds:
                self.__runnerIds[vm_loc] = [str(new_id),]
            else:
                self.__runnerIds[vm_loc].append(str(new_id))
            self.vms.append(new_vm)
            
            try:
                self.resource_checkout(new_vm)
            except:
                log.exception("Unexpected error checking out resources when creating a VM. Programming error?")
                return self.ERROR
            #endfor
            return 0
        
        except Exception, e:
            log.debug("Exception running new instance in StratusLab: %s" %str(e))
            #import traceback
            #traceback.print_exc()
            return -1


    def __cleanKill(self, vm, return_resources=True):
        
        log.verbose('Cleaning caches for VM with id %s' % str(vm.id))
        
        if return_resources and vm.return_resources:
            self.resource_return(vm)
        
        with self.vms_lock:
            
            try:
                self.vms.remove(vm)
            except ValueError:
                log.error("Attempted to remove vm from list that was already removed.")
            
            try:
                self.__runnerIds[StratusLabCluster.__vmMap[vm.id]].remove(vm.id)
            except:
                log.verbose("Attempted to remove already removed runner id")
            
            try:
                del StratusLabCluster.__vmMap[vm.id]        
            except:
                log.verbose("Attempted to remove already removed VM id")


    def __vm_kill(self, vm, runner, clean=False, return_resources=True):
        
        log.debug("Send kill signal to VM %s in StratusLab" % str(vm.id))
        
        try:
            runner.killInstances([int(vm.id)])
        finally:
            if clean:
                self.__cleanKill(vm, return_resources)
    
    
    def vm_destroy(self, vm, return_resources=True, reason=""):
        
        log.debug("Send shutdown signal to VM %s in StratusLab" % str(vm.id))
        
        if len(reason) > 0:
            log.debug("Reason: %s" % reason)
        
        try:
            StratusLabCluster.__idMap[StratusLabCluster.__vmMap[vm.id]].shutdownInstances([int(vm.id)])
            t = threading._Timer(StratusLabCluster.VM_SHUTDOWN, self.__vm_kill, args=[vm, StratusLabCluster.__idMap[StratusLabCluster.__vmMap[vm.id]]])
            log.debug("Waiting in new thread %s to send kill signal to VM %s in StratusLab" % (str(StratusLabCluster.VM_SHUTDOWN), str(vm.id)))
            t.start()
            self.__cleanKill(vm, return_resources)
            return 0      
        except:
            log.debug("VM with id %s shutdown error in StratusLab" % str(vm.id))
            #import traceback
            #traceback.print_exc()
            
            try:
                self.__vm_kill(vm, StratusLabCluster.__idMap[StratusLabCluster.__vmMap[vm.id]], clean=True, return_resources=return_resources)
                log.debug("Managed to kill VM with id %s in StratusLab" % str(vm.id))
                return 0
            except:
                retry = 3
                while retry > 0:
                    log.debug("Error sending kill signal to VM with id %s, %s trials remaining. Retrying..." % (str(vm.id), str(retry)))
                    time.sleep(5)
                    try:
                        self.__vm_kill(vm, StratusLabCluster.__idMap[StratusLabCluster.__vmMap[vm.id]], clean=True, return_resources=return_resources)
                        log.debug("Managed to kill VM with id %s in StratusLab" % str(vm.id))
                        return 0
                    except:
                        pass
                    retry -= 1
                log.debug("No way, can't destroy VM with id %s in StratusLab. Maybe already destroyed, cleaning..." % str(vm.id))
                self.__cleanKill(vm,False)
                #traceback.print_exc()
                return -1


    def vm_poll(self, vm):
        
        try:
            with self.vms_lock:
                new_status = self.VM_STATES.get(StratusLabCluster.__idMap[StratusLabCluster.__vmMap[vm.id]].getVmState(int(vm.id)).upper(), 'Error')
                if vm.status != new_status:
                        vm.last_state_change = int(time.time())
                        log.debug("VM: %s on %s. Changed from %s to %s." % (str(vm.id), self.name, vm.status, new_status))
                        vm.status = new_status
                elif vm.override_status != None and new_status:
                        log.debug("New status for VM %s, override status" % str(vm.id))
                        vm.override_status = None
                        vm.errorconnect = None

            vm.lastpoll = int(time.time())
            return new_status
        except OneException, e:
            log.debug("One exception: %s" % str(e))
            return 'shutdown'
        except Exception, e:
            log.debug("Unknown exception: %s" % str(e))
            return 'unknown'


    def __getstate__(self):
        
        state = self.__dict__.copy()
        del state['vms_lock']
        del state['res_lock']
        return state


    def __setstate__(self, state):
        
        self.__dict__ = state
        self.vms_lock = threading.RLock()
        self.res_lock = threading.RLock()
        self.__setRunnerIds(state['_StratusLabCluster__runnerIds'])


    def __setRunnerIds(self, runnerIds):
        
        self.__runnerIds = runnerIds
        
        for key in self.__runnerIds.keys():
            
            if len(self.__runnerIds[key]) > 0:
                r = Runner(key, StratusLabCluster.__configHolder)
                self.__idMap[key] = r
            
                for runnerId in self.__runnerIds[key]:
                    self.__vmMap[runnerId] = str(key)
                    r.vmIds.append(runnerId)
            else:
                del self.__runnerIds[key]


